import asyncio
import os
import ssl
from datetime import datetime, timedelta
from ssl import Purpose, SSLContext
from typing import Dict, Iterator, Optional

from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePrivateKeyWithSerialization
from cryptography.x509 import Certificate, CertificateSigningRequest

from bxcommon.models.node_type import NodeType
from bxutils import constants, logging
from bxutils.common import url_helper
from bxutils.ssl import ssl_certificate_factory
from bxutils.ssl.data.ssl_file_info import SSLFileInfo
from bxutils.ssl.data.ssl_storage_info import SSLStorageInfo
from bxutils.ssl.extensions import extensions_factory
from bxutils.ssl.ssl_certificate_type import SSLCertificateType

logger = logging.get_logger(__name__)


def is_cert_valid(cert: Certificate, expiration_threshold_days: int = 0) -> bool:
    """
    Checking if a certificate is valid or will still be valid until `expiration_threshold_days` will pass.
    :param cert: the certificate to validate.
    :param expiration_threshold_days: the additional period (in days) in which the certificate should still be valid.
    :return: True if valid, otherwise False.
    """
    utc_now = datetime.utcnow()
    expiration_threshold = timedelta(days=expiration_threshold_days)
    return cert.not_valid_before <= utc_now + expiration_threshold <= cert.not_valid_after


class NodeSSLService:
    """
    A service class for a node that provides - fetching, loading, storing and creating cert chains.
    Pass store_local=False if storing cert the cert chains locally is not required (usually for testing).
    Typical usage is as follows:
        1. Create and load the NodeSSLService.
        2. Check if we should request for a new certificate using the should_renew_node_certificate method.
        3. If required, create a CSR, add it to the node registration request and store the private certificate.
        4. Create an SSLContext by calling create_ssl_context(SSLCertificateType.PRIVATE).
    """
    storage_info: SSLStorageInfo
    certificates: Dict[SSLCertificateType, Certificate]
    private_keys: Dict[SSLCertificateType, EllipticCurvePrivateKeyWithSerialization]
    _store_local: bool
    _node_name: str

    def __init__(self, node_type: NodeType, storage_info: SSLStorageInfo, store_local: bool = True) -> None:
        if storage_info.certificates_info[SSLCertificateType.PRIVATE].key_file_info is None:
            raise ValueError(f"Storage info {storage_info} is missing key file storage info for a private certificate.")
        self.storage_info = storage_info
        self.certificates = {}
        self.private_keys = {}
        self._node_name = node_type.name.lower()
        self._store_local = store_local

    def __repr__(self) -> str:
        return f"{self.__class__.__name__} <{self._node_name}>"

    def blocking_load(self) -> None:
        """
        Load the service in a blocking manner.
        CA and REGISTRATION_ONLY certificates will be fetched from their configured URLs.
        The PRIVATE certificate will be loaded if exists in the local path.
        Local directories will be created if needed.
        :raise RuntimeError: if either the CA certificate or the Node certificate is missing.
        """
        self._make_dirs()
        for cert_type, cert_info in self.storage_info.certificates_info.items():
            cert_file_info = cert_info.cert_file_info
            cert_url = self._get_url(cert_file_info)
            if cert_url is not None:
                logger.trace("Load ssl certificate {}", cert_url)
                cert = ssl_certificate_factory.fetch_cert(cert_url)
                self.certificates[cert_type] = cert
                self._store_cert(cert_type, cert)
            key_file_info = cert_info.key_file_info
            if key_file_info is not None:
                key_url = self._get_url(key_file_info)
                if key_url is not None:
                    logger.trace("Load ssl key {}", key_url)
                    key = ssl_certificate_factory.fetch_key(key_url)
                    self.private_keys[cert_type] = key
                    self._store_key(key, key_file_info)
        if not self.has_valid_certificate(SSLCertificateType.CA):
            raise RuntimeError("Failed to load CA certificate.")
        if (
            not self.has_valid_certificate(SSLCertificateType.REGISTRATION_ONLY)
            and not self.has_valid_certificate(SSLCertificateType.PRIVATE)
        ):
            raise RuntimeError("Failed to load node certificate.")

        logger.info("{} is successfully loaded.", self)

    async def load(self) -> None:
        """
        Load the service in a non blocking manner.
        """
        loop = asyncio.get_event_loop()
        # TODO : convert to proper async
        await loop.run_in_executor(None, self.blocking_load)

    def get_certificate(self, cert_type: SSLCertificateType) -> Certificate:
        """
        Get a certificate from the chain according to a specific type.
        :param cert_type: The certificate type (CA, REGISTRATION_ONLY, PRIVATE) to get.
        :return: an SSL certificate object.
        :raise ValueError: if the requested certificate (cert_type) is missing.
        """
        try:
            return self.certificates[cert_type]
        except KeyError:
            raise ValueError(f"Could not find SSL certificate for type: {cert_type}.")

    def has_valid_certificate(self, cert_type: SSLCertificateType) -> bool:
        """
        Check if the service has a valid certificate for a specific type.
        :param cert_type: The certificate type (CA, REGISTRATION_ONLY, PRIVATE) to check.
        :return: True if has valid certificate, otherwise False.
        """
        return cert_type in self.certificates and is_cert_valid(self.certificates[cert_type])

    def has_invalid_certificate(self, cert_type: SSLCertificateType) -> bool:
        """
        Check if the service has an invalid certificate for a specific type.
        :param cert_type: The certificate type (CA, REGISTRATION_ONLY, PRIVATE) to check.
        :return: True if has a certificate and it is invalid, otherwise False.
        """
        return cert_type in self.certificates and not is_cert_valid(self.certificates[cert_type])

    def should_renew_node_certificate(
            self, expiration_threshold_days: int = constants.DEFAULT_CERTIFICATE_RENEWAL_PERIOD_DAYS
    ) -> bool:
        """
        Check if the service should renew its node private certificate.
        :param expiration_threshold_days: the threshold in days, in which a renewal should be requested.
        :return: True if renewal is required, otherwise False.
        """
        has_cert = SSLCertificateType.PRIVATE in self.certificates \
            and SSLCertificateType.REGISTRATION_ONLY in self.certificates
        if has_cert and is_cert_valid(self.certificates[SSLCertificateType.PRIVATE], expiration_threshold_days):
            private_account = extensions_factory.get_account_id(
                self.certificates[SSLCertificateType.PRIVATE]
            )
            registration_only_account = extensions_factory.get_account_id(
                self.certificates[SSLCertificateType.REGISTRATION_ONLY]
            )
            return private_account != registration_only_account

        return True

    def blocking_store_node_certificate(self, cert: Certificate) -> None:
        """
        Store a node private certificate in a blocking manner
        :param cert: the node private certificate to be stored.
        """
        key = self.private_keys[SSLCertificateType.PRIVATE]
        if self._store_local:
            ssl_folder_path = self.storage_info.ssl_folder_path
            key_file_info = self.storage_info.certificates_info[SSLCertificateType.PRIVATE].key_file_info
            key_folder_path = os.path.join(ssl_folder_path, key_file_info.sub_directory_name)  # pyre-ignore
            ssl_certificate_factory.store_key(key, key_folder_path, self._node_name)
        self.certificates[SSLCertificateType.PRIVATE] = cert
        self._store_cert(SSLCertificateType.PRIVATE, cert)
        node_id = extensions_factory.get_node_id(cert)
        logger.debug("{} successfully stored private SSL certificate: {} ({}).", self, cert, node_id)

    async def store_node_certificate(self, cert: Certificate) -> None:
        """
        Store a node private certificate in a non blocking manner
        :param cert: the node private certificate to be stored.
        """
        loop = asyncio.get_event_loop()
        # TODO : convert to proper async
        await loop.run_in_executor(None, self.blocking_store_node_certificate, cert)

    def create_csr(self) -> CertificateSigningRequest:
        """
        Create a Certificate Signing Request (CSR) for a private node certificate.
        :return: CSR (https://en.wikipedia.org/wiki/Certificate_signing_request)
        """
        key = ssl_certificate_factory.generate_key()
        self.private_keys[SSLCertificateType.PRIVATE] = key
        template_cert = self.certificates[SSLCertificateType.REGISTRATION_ONLY]
        csr = ssl_certificate_factory.create_csr(key, template_cert)
        logger.debug("{} successfully created private SSL CSR: {}.", self, csr)
        return csr

    def create_ssl_context(self, cert_type: SSLCertificateType) -> SSLContext:
        """
        Create an SSL context object to be used for wrapping SSLSockets
        (https://docs.python.org/3.7/library/ssl.html#ssl.SSLSocket).
        :param cert_type: the certificate type to use when creating the context.
        :return: an SSL context object (https://docs.python.org/3.7/library/ssl.html#ssl.SSLContext)
        """
        cert_info = self.storage_info.certificates_info[cert_type]
        context = ssl.create_default_context(
            Purpose.SERVER_AUTH,
            cafile=self._get_file_path(self.storage_info.certificates_info[SSLCertificateType.CA].cert_file_info)
        )
        context.load_cert_chain(
            self._get_file_path(cert_info.cert_file_info), self._get_file_path(cert_info.key_file_info)  # pyre-ignore
        )
        context.check_hostname = False
        logger.debug(
            "{} successfully created SSL context for certificate: {}.",
            self,
            self.certificates[cert_type]
        )
        return context

    def get_node_id(self) -> str:
        cert = self.get_certificate(SSLCertificateType.PRIVATE)
        node_id = extensions_factory.get_node_id(cert)
        if node_id is None:
            raise TypeError(f"Node id is missing in private certificate: {cert}!")
        return node_id

    def get_account_id(self) -> Optional[str]:
        cert = None
        try:
            cert = self.get_certificate(SSLCertificateType.REGISTRATION_ONLY)
        except ValueError:
            pass
        if cert is None:
            logger.info("Could not find SSL certificate, continue without account settings")
            return None
        account_id = extensions_factory.get_account_id(cert)
        if account_id is None:
            logger.info("Account id is missing in certificate: {}!", cert)
        return account_id

    def _store_cert(self, cert_type: SSLCertificateType, cert: Certificate) -> None:
        if not self._store_local:
            logger.debug("{} skip saving {} to local storage.", self, cert)
            return
        cert_file_info = self.storage_info.certificates_info[cert_type].cert_file_info
        cert_folder_path = os.path.join(
            self.storage_info.ssl_folder_path, cert_file_info.sub_directory_name
        )
        cert_file_name = cert_type.name.lower() if cert_type == SSLCertificateType.CA else self._node_name
        ssl_certificate_factory.store_certificate(cert, cert_folder_path, cert_file_name)

    def _store_key(self, key: EllipticCurvePrivateKeyWithSerialization, file_info: Optional[SSLFileInfo]) -> None:
        if not self._store_local or file_info is None:
            logger.debug("{} skip saving private key to local storage.", self)
            return
        ssl_folder_path = self.storage_info.ssl_folder_path
        folder_path = os.path.join(ssl_folder_path, file_info.sub_directory_name)
        file_parts = file_info.file_name.split("_")[:-1]
        file_prefix = "_".join(file_parts)
        ssl_certificate_factory.store_key(key, folder_path, file_prefix)

    def _make_dirs(self) -> None:
        if not self._store_local:
            return
        ssl_folder_path = self.storage_info.ssl_folder_path

        def _get_dirs() -> Iterator[str]:
            yield ssl_folder_path
            for cert_info in self.storage_info.certificates_info.values():
                yield os.path.join(ssl_folder_path, cert_info.cert_file_info.sub_directory_name)
        for folder_path in _get_dirs():
            if not os.path.exists(folder_path):
                os.makedirs(folder_path)
                logger.debug("{} creating missing directory: {}.", self, folder_path)

    def _get_url(self, file_info: SSLFileInfo) -> Optional[str]:
        url = file_info.url
        file_path = self._get_file_path(file_info)
        if url is None and self._store_local and os.path.exists(file_path):
            return url_helper.url_join("file:", file_path)
        else:
            return url

    def _get_file_path(self, file_info: SSLFileInfo) -> str:
        ssl_folder_path = self.storage_info.ssl_folder_path
        return os.path.join(ssl_folder_path, file_info.sub_directory_name, file_info.file_name)



from bxcommon.models.blockchain_protocol import BlockchainProtocol
from bxgateway.connections.btc.btc_gateway_node import BtcGatewayNode
from bxgateway.connections.eth.eth_gateway_node import EthGatewayNode
from bxutils import logging

logger = logging.get_logger(__name__)


def get_gateway_node_type(blockchain_protocol):
    # TODO: This is temporary logic that will be replaced with list of valid protocols and networks from SDN
    if blockchain_protocol == BlockchainProtocol.ETHEREUM.value:
        return EthGatewayNode

    if blockchain_protocol == BlockchainProtocol.ONTOLOGY.value:
        from bxgateway.connections.ont.ont_gateway_node import OntGatewayNode
        return OntGatewayNode

    return BtcGatewayNode



import asyncio
import base64
import uuid
import json
import sys
import shlex
import os
from enum import Enum
from asyncio import StreamReader, StreamReaderProtocol, StreamWriter, CancelledError
from asyncio.streams import FlowControlMixin
from argparse import ArgumentParser, Namespace, RawDescriptionHelpFormatter
from json import JSONDecodeError
from typing import Optional, Union, Dict, List, Any
from aiohttp import ClientSession, ClientResponse, ClientConnectorError, ContentTypeError, ClientConnectionError

"""
Valid import path: bxcommon.models, bxcommon.rpc
"""
from bxcommon.models.blockchain_protocol import BlockchainProtocol
from bxcommon.rpc import rpc_constants
from bxcommon.rpc.rpc_request_type import RpcRequestType

COMMANDS_HELP = [
    "{:<18} exit the CLI.".format("exit"),
    "{:<18} print detailed help.".format("help"),
    "{:<18} send transaction to the bloXroute BDN.".format("blxr_tx"),
    "{:<18} get the status of the bloXroute Gateway.".format("gateway_status"),
    "{:<18} get the memory stats of the bloXroute Gateway.".format("memory"),
    "{:<18} shutdown the Gateway server.".format("stop"),
    "{:<18} get the bloXroute Gateway connected peers info.".format("peers"),
    "{:<18} get BDN performance stats.".format("bdn_performance"),
    "{:<18} get quota usage status.".format("quota_usage"),
    "{:<18} dump transaction service to file.".format("tx_service"),
    "{:<18} get status of transaction.".format("tx_status")
]


class CLICommand(Enum):
    EXIT = 0
    HELP = 1


class ArgParserFile:

    def __init__(self, writer: StreamWriter):
        self._writer = writer

    def write(self, message: str) -> int:
        lines = message.split("\n")
        for line in lines:
            self._writer.write(line.encode("utf-8"))
            self._writer.write(b"\n")
        return len(message)


class RpcClient:

    def __init__(
        self,
        rpc_host: str,
        rpc_port: int,
        rpc_user: str,
        rpc_password: str,
        account_id: str,
        secret_hash: str,
        cloud_api_url: str,
        target_is_cloud_api: bool,
        interactive: bool = False
    ):
        self._session = ClientSession()
        self.target_is_cloud_api = target_is_cloud_api
        self.interactive = interactive
        if target_is_cloud_api:
            self._rpc_url = cloud_api_url
            self._encoded_auth = base64.b64encode(f"{account_id}:{secret_hash}".encode("utf-8")).decode("utf-8")
        else:
            self._rpc_url = f"http://{rpc_host}:{rpc_port}/"
            self._encoded_auth = base64.b64encode(f"{rpc_user}:{rpc_password}".encode("utf-8")).decode("utf-8")

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    async def make_request(
        self,
        method: RpcRequestType,
        request_id: Optional[str] = None,
        request_params: Union[Dict[str, Any], None] = None
    ) -> Optional[ClientResponse]:
        if request_id is None:
            request_id = str(uuid.uuid4())
        json_data = {
            "method": method.name.lower(),
            "id": request_id,
            "params": request_params
        }
        synchronous = True
        if request_params:
            synchronous = request_params.get(rpc_constants.SYNCHRONOUS_PARAMS_KEY, True)

        if not synchronous and self.interactive:
            asyncio.create_task(
                self._session.post(
                    self._rpc_url,
                    data=json.dumps(json_data),
                    headers={
                        rpc_constants.CONTENT_TYPE_HEADER_KEY: rpc_constants.PLAIN_HEADER_TYPE,
                        rpc_constants.AUTHORIZATION_HEADER_KEY: self._encoded_auth
                    }
                )
            )
            return None

        return await self._session.post(
            self._rpc_url,
            data=json.dumps(json_data),
            headers={
                rpc_constants.CONTENT_TYPE_HEADER_KEY: rpc_constants.PLAIN_HEADER_TYPE,
                rpc_constants.AUTHORIZATION_HEADER_KEY: self._encoded_auth
            }
        )

    async def get_server_help(self) -> ClientResponse:
        return await self._session.get(
            self._rpc_url,
            headers={
                rpc_constants.CONTENT_TYPE_HEADER_KEY: rpc_constants.PLAIN_HEADER_TYPE,
                rpc_constants.AUTHORIZATION_HEADER_KEY: self._encoded_auth
            }
        )

    async def close(self) -> None:
        await self._session.close()


def merge_params(opts: Namespace, unrecognized_params: List[str]) -> Namespace:
    merged_opts = Namespace()
    merged_opts.__dict__ = opts.__dict__.copy()
    if merged_opts.command == RpcRequestType.BLXR_TX:
        if merged_opts.request_params is None and unrecognized_params:
            transaction_payload = unrecognized_params[0]
            merged_opts.request_params = {
                rpc_constants.TRANSACTION_PARAMS_KEY: transaction_payload,
                rpc_constants.SYNCHRONOUS_PARAMS_KEY: True,
                rpc_constants.STATUS_TRACK_PARAMS_KEY: rpc_constants.STATUS_TRACK_PARAMS_KEY in unrecognized_params,
                rpc_constants.NONCE_MONITORING_PARAMS_KEY: rpc_constants.NONCE_MONITORING_PARAMS_KEY in unrecognized_params,
                rpc_constants.ACCOUNT_ID_PARAMS_KEY: opts.account_id if opts.cloud_api else opts.rpc_user,
                rpc_constants.BLOCKCHAIN_PROTOCOL_PARAMS_KEY: opts.blockchain_protocol,
                rpc_constants.BLOCKCHAIN_NETWORK_PARAMS_KEY: opts.blockchain_network
            }
    elif merged_opts.command == RpcRequestType.GATEWAY_STATUS:
        if merged_opts.request_params is None and unrecognized_params:
            merged_opts.request_params = {rpc_constants.DETAILS_LEVEL_PARAMS_KEY: unrecognized_params[0]}
    elif merged_opts.command == RpcRequestType.QUOTA_USAGE:
        if merged_opts.request_params is None:
            merged_opts.request_params = {
                rpc_constants.ACCOUNT_ID_PARAMS_KEY: opts.account_id if opts.cloud_api else opts.rpc_user
            }
    elif merged_opts.command == RpcRequestType.TX_SERVICE:
        if merged_opts.request_params is None and unrecognized_params:
            merged_opts.request_params = {rpc_constants.TX_SERVICE_FILE_NAME_PARAMS_KEY: unrecognized_params[0]}
    elif merged_opts.command == RpcRequestType.ADD_BLOCKCHAIN_PEER or \
            merged_opts.command == RpcRequestType.REMOVE_BLOCKCHAIN_PEER:
        if merged_opts.request_params is None and unrecognized_params:
            merged_opts.request_params = {rpc_constants.BLOCKCHAIN_PEER_PARAMS_KEY: unrecognized_params[0]}

    return merged_opts


async def format_response(response: ClientResponse) -> str:
    try:
        response_json_str = await response.text()
        response_json_dict = json.loads(json.loads(response_json_str))
        result = response_json_dict.get("result", None)
        if not result:
            result = response_json_dict.get("error", "Unknown error")
        return json.dumps(result, indent=4)
    except (JSONDecodeError, ContentTypeError, TypeError):
        return await response.text()


async def handle_command(opts: Namespace, client: RpcClient, stdout_writer: StreamWriter) -> None:
    if opts.debug:
        stdout_writer.write(f"executing with opts: {opts}\n".encode("utf-8"))
    response_text: Optional[str] = None
    if isinstance(opts.command, RpcRequestType):
        response: Optional[ClientResponse] = await client.make_request(
            opts.command, opts.request_id, opts.request_params
        )
        if response:
            response_text = await format_response(response)
        else:
            response_text = "{'message': 'Sent async request. Not waiting for response'}"
    elif opts.command == CLICommand.HELP:
        try:
            response: ClientResponse = await client.get_server_help()
            response_text = "Server Help:\n\n{}".format(await format_response(response))
        except (ClientConnectorError, ClientConnectionError):
            pass
    if response_text is not None:
        stdout_writer.write(response_text.encode("utf-8").decode("unicode_escape").encode("utf-8"))
        stdout_writer.write(b"\n")
        await stdout_writer.drain()
    await asyncio.sleep(0)


async def parse_and_handle_command(
    arg_parser: ArgumentParser,
    stdout_writer: StreamWriter,
    client: RpcClient,
    shell_args: Optional[List[str]] = None
) -> None:
    try:
        opts, params = arg_parser.parse_known_args(shell_args)
    except KeyError as unrecognized_command:
        stdout_writer.write(
            f"unrecognized command {str(unrecognized_command).lower()} entered, ignoring!\n".encode("utf-8")
        )
        await stdout_writer.drain()
    else:
        opts = merge_params(opts, params)
        await handle_command(opts, client, stdout_writer)


async def run_cli(
    rpc_host: str,
    rpc_port: int,
    rpc_user: str,
    rpc_password: str,
    account_id: str,
    secret_hash: str,
    cloud_api_url: str,
    target_is_cloud_api: bool,
    arg_parser: ArgumentParser,
    stdin_reader: StreamReader,
    stdout_writer: StreamWriter

) -> None:

    interactive = True
    async with RpcClient(
        rpc_host, rpc_port, rpc_user, rpc_password, account_id, secret_hash, cloud_api_url, target_is_cloud_api,
        interactive
    ) as client:
        while True:
            stdout_writer.write(b">> ")
            await stdout_writer.drain()
            line = (await stdin_reader.readline()).decode("utf-8")
            if line == "\n":
                await asyncio.sleep(0)
                continue
            if "exit" in line.lower():
                break
            shell_args = shlex.split(line)
            if len(shell_args) == 0:
                shell_args.append("--interactive-shell")
            if "-h" in shell_args or "--help" in shell_args or "help" in shell_args:
                arg_parser.print_help(file=ArgParserFile(stdout_writer))  # pyre-ignore
            await parse_and_handle_command(arg_parser, stdout_writer, client, shell_args)


def get_command(command: str) -> Union[CLICommand, RpcRequestType]:
    command = command.upper()
    if command in CLICommand.__members__:
        return CLICommand[command]
    else:
        return RpcRequestType[command]


def get_command_help() -> str:
    commands = [command.lower() for command in CLICommand.__members__.keys()] + \
               [command.lower() for command in RpcRequestType.__members__.keys()]
    return f"The CLI command (valid values: {commands})."


def get_description() -> str:
    desc = "Commands:\n"
    for command_help in COMMANDS_HELP:
        desc = f"{desc}\n{command_help}"
    return desc


def add_run_arguments(arg_parser: ArgumentParser) -> None:
    arg_parser.add_argument(
        "command",
        help=get_command_help(),
        type=get_command
    )
    arg_parser.add_argument(
        "--request-id",
        help="The Gateway RPC request unique identifier (default: {}).".format(None),
        type=str,
        default=None
    )
    arg_parser.add_argument(
        "--request-params",
        help="The Gateway RPC request params (default: {}).".format(None),
        type=json.loads,
        default=None
    )
    arg_parser.add_argument(
        "--blockchain-protocol",
        help="Blockchain protocol. e.g Ethereum, BitcoinCash. (default: Ethereum)",
        type=str,
        default=BlockchainProtocol.ETHEREUM.name
    )
    arg_parser.add_argument(
        "--blockchain-network",
        help="Blockchain network. e.g Mainnet, Ropsten, Rinkeby, and Testnet. (default: Mainnet)",
        type=str,
        default=rpc_constants.MAINNET_NETWORK_NAME
    )


def add_base_arguments(arg_parser: ArgumentParser) -> None:
    arg_parser.add_argument(
        "--cloud-api",
        help=f"Should the bloxroute-cli send the request to bloXroute Cloud API",
        action="store_true"
    )
    arg_parser.add_argument(
        "--cloud-api-url",
        help=f"bloXroute's cloud-api DNS name (default: {rpc_constants.CLOUD_API_URL})",
        type=str,
        default=rpc_constants.CLOUD_API_URL
    )
    arg_parser.add_argument(
        "--account-id",
        help=f"The account's ID. Contact support@bloxroute.com for assistance. ",
        type=str
    )
    arg_parser.add_argument(
        "--secret-hash",
        help=f"The account's secret key. Contact support@bloxroute.com for assistance. ",
        type=str,
    )
    arg_parser.add_argument(
        "--rpc-host",
        help="The Gateway RPC host (default: {}).".format(rpc_constants.DEFAULT_RPC_HOST),
        type=str,
        default=rpc_constants.DEFAULT_RPC_HOST
    )
    arg_parser.add_argument(
        "--rpc-port",
        help="The Gateway RPC port (default: {}).".format(rpc_constants.DEFAULT_RPC_PORT),
        type=int,
        default=rpc_constants.DEFAULT_RPC_PORT
    )
    arg_parser.add_argument(
        "--rpc-user",
        help=f"The Gateway's RPC server username. Contact support@bloxroute.com for assistance."
             f"(default: {rpc_constants.DEFAULT_RPC_USER})",
        type=str,
        default=rpc_constants.DEFAULT_RPC_USER
    )
    arg_parser.add_argument(
        "--rpc-password",
        help=f"The Gateway's RPC server password. (default: {rpc_constants.DEFAULT_RPC_PASSWORD})",
        type=str,
        default=rpc_constants.DEFAULT_RPC_PASSWORD
    )
    arg_parser.add_argument(
        "--interactive-shell",
        help="Run the Gateway RPC client in interactive Shell mode (default: False)",
        action="store_true",
        default=False
    )
    arg_parser.add_argument(
        "--debug",
        help="Run the CLI in debug mode (default: False)",
        action="store_true",
        default=False
    )


def validate_args(opts: Namespace, stdout_writer: StreamWriter) -> bool:
    if ("account_id" in opts and opts.account_id is not None) and \
            ("secret_hash" in opts and opts.secret_hash is not None):
        opts.cloud_api = True
        sys.argv.append("--cloud-api")
        return True

    elif ("cloud_api" in opts and opts.cloud_api is True):
        if ("cloud_api_url" not in opts or ("cloud_api_url" in opts and opts.cloud_api_url is None)) or \
           ("account_id" not in opts or ("account_id" in opts and opts.account_id is None)) or \
           ("secret_hash" not in opts or ("secret_hash" in opts and opts.secret_hash is None)):
            stdout_writer.write("The use of bloXroute Cloud API requires the following two arguments: "
                                "--account-id, --secret-hash.\n".encode("utf-8"))
            return False

    elif ("rpc_host" not in opts or ("rpc_host" in opts and opts.rpc_host is None)) or \
         ("rpc_port" not in opts or ("rpc_port" in opts and opts.rpc_port is None)) or \
         ("rpc_user" not in opts or ("rpc_user" in opts and opts.rpc_user is None)) or \
         ("rpc_password" not in opts or ("rpc_password" in opts and opts.rpc_password is None)):
            stdout_writer.write("The use of bloXroute Gateway requires the following two arguments: "
                                "--rpc-user, --rpc-password.\n".encode("utf-8"))
            return False

    return True


async def main():
    arg_parser = ArgumentParser(
        prog="bloXroute CLI", epilog=get_description(), formatter_class=RawDescriptionHelpFormatter
    )
    cli_parser = ArgumentParser(
        prog="bloXroute CLI", epilog=get_description(), formatter_class=RawDescriptionHelpFormatter
    )
    add_base_arguments(arg_parser)
    add_base_arguments(cli_parser)
    add_run_arguments(cli_parser)
    args = sys.argv[1:]
    if "-h" in args or "--help" in args:
        cli_parser.print_help()
        arg_parser.exit()

    opts, params = arg_parser.parse_known_args()
    stdin_reader = StreamReader()
    cli_loop = asyncio.get_event_loop()
    await cli_loop.connect_read_pipe(
        lambda: StreamReaderProtocol(stdin_reader),
        sys.stdin
    )
    transport, protocol = await cli_loop.connect_write_pipe(
        lambda: FlowControlMixin(), os.fdopen(sys.stdout.fileno(), "wb")
    )
    stdout_writer = StreamWriter(transport, protocol, None, cli_loop)
    if not validate_args(opts, stdout_writer):
        exit(1)

    try:

        if opts.interactive_shell or len(params) == 0:
            await run_cli(
                opts.rpc_host, opts.rpc_port, opts.rpc_user, opts.rpc_password,
                opts.account_id, opts.secret_hash, opts.cloud_api_url, opts.cloud_api,
                cli_parser, stdin_reader, stdout_writer
            )
        else:
            if "help" in sys.argv:
                cli_parser.print_help(file=ArgParserFile(stdout_writer))
            async with RpcClient(
                opts.rpc_host, opts.rpc_port, opts.rpc_user, opts.rpc_password,
                opts.account_id, opts.secret_hash, opts.cloud_api_url, opts.cloud_api
            ) as client:
                await parse_and_handle_command(cli_parser, stdout_writer, client)
    except (ClientConnectorError, ClientConnectionError, TimeoutError, CancelledError) as e:
        stdout_writer.write(f"Connection to RPC server is broken: {e}, exiting!\n".encode("utf-8"))
        await stdout_writer.drain()
        if opts.debug:
            raise
    finally:
        stdout_writer.close()


def run_main():
    loop = asyncio.get_event_loop()
    loop.run_until_complete(main())


if __name__ == "__main__":
    run_main()



import asyncio
from asyncio import Future
from typing import Optional, Union, cast

from aiohttp import WSMessage
from aiohttp.web import Request
from aiohttp.web_ws import WebSocketResponse
from websockets import WebSocketServerProtocol

from bxcommon.rpc.abstract_ws_rpc_handler import AbstractWsRpcHandler, WsRequest
from bxcommon.rpc.rpc_errors import RpcError
from bxcommon.rpc.json_rpc_response import JsonRpcResponse


class WsConnection:
    def __init__(
        self,
        websocket: Union[WebSocketResponse, WebSocketServerProtocol],
        path: str,
        ws_rpc_handler: AbstractWsRpcHandler,
    ) -> None:
        self.ws = websocket
        self.path = path  # currently unused
        self.ws_rpc_handler = ws_rpc_handler

        self.request_handler: Optional[Future] = None
        self.publish_handler: Optional[Future] = None
        self.alive_handler: Optional[Future] = None

    async def handle(self, request: Request) -> WebSocketResponse:
        request_handler = asyncio.ensure_future(self.handle_request(request))
        publish_handler = asyncio.ensure_future(self.handle_publications())
        alive_handler = asyncio.ensure_future(self.ws_rpc_handler.wait_for_close())

        self.request_handler = request_handler
        self.publish_handler = publish_handler
        self.alive_handler = alive_handler

        await asyncio.wait(
            [request_handler, publish_handler, alive_handler], return_when=asyncio.FIRST_COMPLETED
        )
        return self.ws

    async def handle_request(self, request: Request) -> None:
        websocket = self.ws
        await websocket.prepare(request)
        async for message in websocket:
            try:
                response = await self.ws_rpc_handler.handle_request(
                    # pyre-ignore[6] Expected `multidict.CIMultiDictProxy[typing.Any]`
                    WsRequest(cast(WSMessage, message), request.headers)
                )
            except RpcError as err:
                response = JsonRpcResponse(err.id, error=err).to_jsons()
            await websocket.send_str(response)

    async def handle_publications(self, ) -> None:
        websocket = self.ws
        while not websocket.closed:
            message = await self.ws_rpc_handler.get_next_subscribed_message()
            await websocket.send_bytes(
                self.ws_rpc_handler.serialize_cached_subscription_message(message)
            )

    async def close(self) -> None:
        self.ws_rpc_handler.close()

        request_handler = self.request_handler
        if request_handler is not None:
            request_handler.cancel()

        publish_handler = self.publish_handler
        if publish_handler is not None:
            publish_handler.cancel()

        alive_handler = self.alive_handler
        if alive_handler is not None:
            alive_handler.cancel()

        # cleanup to avoid circular reference and allow immediate GC.
        self.request_handler = None
        self.publish_handler = None
        self.alive_handler = None

        await self.ws.close()



import datetime
from typing import Optional

from bxcommon.rpc.bx_json_rpc_request import BxJsonRpcRequest
from bxcommon.rpc.json_rpc_response import JsonRpcResponse
from bxcommon.rpc.requests.abstract_rpc_request import AbstractRpcRequest, Node
from bxcommon.rpc.rpc_errors import RpcInvalidParams
from bxcommon.utils.object_hash import Sha256Hash


TRANSACTION_HASH_KEY = "transaction_hash"
NETWORK_NUM_KEY = "network_num"


class TransactionStatusRpcRequest(AbstractRpcRequest[Node]):
    help = {
        "params": f"[Required - {TRANSACTION_HASH_KEY}: Transaction hash to check status of] ",
        "description": "Reports status of transaction: "
                       "time received, short id assigned, etc."
    }

    def __init__(self, request: BxJsonRpcRequest, node: Node):
        self.transaction_hash: Optional[Sha256Hash] = None
        self.network_num: Optional[int] = None

        super().__init__(request, node)

    def validate_params(self) -> None:
        params = self.params
        if not isinstance(params, dict):
            raise RpcInvalidParams(
                self.request_id,
                "Params request field is either missing or not a dictionary type."
            )

        if TRANSACTION_HASH_KEY not in params:
            raise RpcInvalidParams(
                self.request_id,
                "Transaction hash was missing from RPC params."
            )

        transaction_hash_str = params[TRANSACTION_HASH_KEY]
        try:
            transaction_hash = Sha256Hash.from_string(transaction_hash_str)
        except Exception as _e:
            raise RpcInvalidParams(
                self.request_id,
                f"Invalid transaction hash: {transaction_hash_str}"
            )
        else:
            self.transaction_hash = transaction_hash

    async def process_request(self) -> JsonRpcResponse:
        transaction_hash = self.transaction_hash
        assert transaction_hash is not None

        transaction_service = self.node.get_tx_service(self.get_network_num())
        transaction_key = transaction_service.get_transaction_key(transaction_hash)

        status = "unknown"
        short_ids = []
        assigned_time = "n/a"
        if transaction_service.has_transaction_contents_by_key(transaction_key):
            status = "pending short ID"

        if transaction_service.has_transaction_short_id_by_key(transaction_key):
            status = "assigned short ID"
            short_ids = transaction_service.get_short_ids_by_key(transaction_key)
            most_recent_timestamp = max([
                transaction_service.get_short_id_assign_time(short_id)
                for short_id in short_ids
            ])
            assigned_time = datetime.datetime.fromtimestamp(most_recent_timestamp).isoformat()

        payload = {
            "status": status,
            "short_ids": short_ids,
            "assignment_time": assigned_time
        }
        return JsonRpcResponse(self.request_id, payload)

    def get_network_num(self) -> Optional[int]:
        return self.network_num



from typing import Type

from bxcommon.messages.abstract_message import AbstractMessage
from bxcommon.messages.abstract_message_factory import AbstractMessageFactory
from bxgateway.messages.ont.addr_ont_message import AddrOntMessage
from bxgateway.messages.ont.block_ont_message import BlockOntMessage
from bxgateway.messages.ont.consensus_ont_message import OntConsensusMessage
from bxgateway.messages.ont.get_addr_ont_message import GetAddrOntMessage
from bxgateway.messages.ont.get_blocks_ont_message import GetBlocksOntMessage
from bxgateway.messages.ont.get_data_ont_message import GetDataOntMessage
from bxgateway.messages.ont.get_headers_ont_message import GetHeadersOntMessage
from bxgateway.messages.ont.headers_ont_message import HeadersOntMessage
from bxgateway.messages.ont.inventory_ont_message import InvOntMessage
from bxgateway.messages.ont.notfound_ont_message import NotFoundOntMessage
from bxgateway.messages.ont.ont_message import OntMessage
from bxgateway.messages.ont.ont_message_type import OntMessageType
from bxgateway.messages.ont.ping_ont_message import PingOntMessage
from bxgateway.messages.ont.pong_ont_message import PongOntMessage
from bxgateway.messages.ont.tx_ont_message import TxOntMessage
from bxgateway.messages.ont.ver_ack_ont_message import VerAckOntMessage
from bxgateway.messages.ont.version_ont_message import VersionOntMessage


class _OntMessageFactory(AbstractMessageFactory):
    _MESSAGE_TYPE_MAPPING = {
        OntMessageType.VERSION: VersionOntMessage,
        OntMessageType.VERACK: VerAckOntMessage,
        OntMessageType.GET_ADDRESS: GetAddrOntMessage,
        OntMessageType.ADDRESS: AddrOntMessage,
        OntMessageType.PING: PingOntMessage,
        OntMessageType.PONG: PongOntMessage,
        OntMessageType.CONSENSUS: OntConsensusMessage,
        OntMessageType.INVENTORY: InvOntMessage,
        OntMessageType.GET_DATA: GetDataOntMessage,
        OntMessageType.GET_HEADERS: GetHeadersOntMessage,
        OntMessageType.GET_BLOCKS: GetBlocksOntMessage,
        OntMessageType.BLOCK: BlockOntMessage,
        OntMessageType.HEADERS: HeadersOntMessage,
        OntMessageType.TRANSACTIONS: TxOntMessage,
        OntMessageType.NOT_FOUND: NotFoundOntMessage
    }

    def __init__(self):
        super(_OntMessageFactory, self).__init__()
        self.message_type_mapping = self._MESSAGE_TYPE_MAPPING

    def get_base_message_type(self) -> Type[AbstractMessage]:
        return OntMessage


ont_message_factory = _OntMessageFactory()



import time
from datetime import datetime, timedelta
from datetime import time as dttime
from typing import TypeVar, Optional, Callable, Any

from bxutils import constants

T = TypeVar("T")
R = TypeVar("R")


def fibonacci(sequence_number: int) -> int:
    return int(
        (constants.FIBONACCI_GOLDEN_RATIO ** sequence_number -
         (1 - constants.FIBONACCI_GOLDEN_RATIO) ** sequence_number) / 5 ** .5
    )


def optional_map(val: Optional[T], mapper: Callable[[T], R]) -> Optional[R]:
    if val is None:
        return val
    else:
        return mapper(val)


def or_else(val: Optional[T], default: T) -> T:
    if val is None:
        return default
    else:
        return val


def identity(val: T) -> T:
    return val


def time_to_date_str(epoch_time: int) -> str:
    return time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(epoch_time))


def seconds_until_eod() -> float:
    dt = datetime.now()
    tomorrow = dt + timedelta(days=1)
    timedelta_until_eod = datetime.combine(tomorrow, dttime.min) - dt
    return timedelta_until_eod.total_seconds()


def memoize(fn: Callable) -> Callable:
    cache = {}

    async def memo_fn(*args):
        if args not in cache:
            cache[args] = await fn(*args)
        return cache[args]

    return memo_fn


class LazyBind:
    def __init__(self) -> None:
        self.bound_fn: Optional[Callable[..., Any]] = None

    def bind(self, bound_fn: Callable[..., Any]) -> None:
        self.bound_fn = bound_fn

    def call(self, *args, **kwargs) -> Any:
        bound_fn = self.bound_fn
        assert bound_fn is not None
        return bound_fn(*args, **kwargs)


def is_greater_or_eq_version(version1: str, version2: str) -> bool:
    """
    :param version1: version number ex) "v1.61.13"
    :param version2: version number ex) "v1.61.12"
    :return: True if version1 >= version2 else False
    """
    if version1 is None or version2 is None:
        return True
    try:
        v1_list = [int(x) for x in version1.strip("v").split(".")]
        v2_list = [int(x) for x in version2.strip("v").split(".")]
        for i in range(min(len(v1_list), len(v2_list))):
            if v1_list[i] > v2_list[i]:
                return True
            elif v1_list[i] < v2_list[i]:
                return False
        return True
    except ValueError:
        return False


def safe_divide(value1: float, value2: float) -> float:
    if value2 == 0:
        return 0

    return value1 / value2


def bind_range(min_value: float, max_value: float, value: float) -> float:
    return min(max_value, max(min_value, value))



from abc import abstractmethod
from typing import Type, List

from bxgateway.gateway_opts import GatewayOpts
from bxgateway.testing import gateway_helpers
from bxcommon import constants
from bxcommon.messages.abstract_message import AbstractMessage
from bxcommon.test_utils import helpers
from bxcommon.test_utils.abstract_test_case import AbstractTestCase
from bxgateway.connections.abstract_gateway_node import AbstractGatewayNode
from bxgateway.testing.mocks import full_gateway_node


class AbstractGatewayIntegrationTest(AbstractTestCase):
    """
    Test scaffolding for a two gateway network, with connections to a simulated relay and blockchain connection.
    """

    def setUp(self) -> None:
        if constants.USE_EXTENSION_MODULES:
            helpers.set_extensions_parallelism()

        self.gateway_info_1 = full_gateway_node.of_type(self.gateway_class(), self.gateway_opts_1())
        self.gateway_info_2 = full_gateway_node.of_type(self.gateway_class(), self.gateway_opts_2())

        (
            self.gateway_1,
            self.blockchain_fileno_1,
            self.relay_fileno_1,
            self.gateway_fileno_1,
            _,
            _,
            _,
            self.blockchain_connection_1,
            self.gateway_connection_1,
            self.relay_connection_1,
        ) = self.gateway_info_1

        (
            self.gateway_2,
            self.blockchain_fileno_2,
            self.relay_fileno_2,
            self.gateway_fileno_2,
            _,
            _,
            _,
            self.blockchain_connection_2,
            self.gateway_connection_2,
            self.relay_connection_2,
        ) = self.gateway_info_2

    @abstractmethod
    def gateway_class(self) -> Type[AbstractGatewayNode]:
        pass

    @abstractmethod
    def gateway_opts_1(self) -> GatewayOpts:
        return gateway_helpers.get_gateway_opts(9000)

    @abstractmethod
    def gateway_opts_2(self) -> GatewayOpts:
        return gateway_helpers.get_gateway_opts(9001)

    def gateway_1_receive_message_from_blockchain(self, message: AbstractMessage):
        helpers.receive_node_message(self.gateway_1, self.blockchain_fileno_1, message.rawbytes())

    def gateway_1_receive_message_from_relay(self, message: AbstractMessage):
        helpers.receive_node_message(self.gateway_1, self.relay_fileno_1, message.rawbytes())

    def gateway_1_receive_message_from_gateway(self, message: AbstractMessage):
        helpers.receive_node_message(self.gateway_1, self.gateway_fileno_1, message.rawbytes())

    def gateway_2_receive_message_from_blockchain(self, message: AbstractMessage):
        helpers.receive_node_message(self.gateway_2, self.blockchain_fileno_2, message.rawbytes())

    def gateway_2_receive_message_from_relay(self, message: AbstractMessage):
        helpers.receive_node_message(self.gateway_2, self.relay_fileno_2, message.rawbytes())

    def gateway_2_receive_message_from_gateway(self, message: AbstractMessage):
        helpers.receive_node_message(self.gateway_2, self.gateway_fileno_2, message.rawbytes())

    def gateway_1_get_queued_messages_for_blockchain(self) -> List[AbstractMessage]:
        return helpers.get_queued_node_messages(self.gateway_1, self.blockchain_fileno_1)

    def gateway_1_get_queued_messages_for_relay(self) -> List[AbstractMessage]:
        return helpers.get_queued_node_messages(self.gateway_1, self.relay_fileno_1)

    def gateway_1_get_queued_messages_for_gateway(self) -> List[AbstractMessage]:
        return helpers.get_queued_node_messages(self.gateway_1, self.gateway_fileno_1)

    def gateway_2_get_queued_messages_for_blockchain(self) -> List[AbstractMessage]:
        return helpers.get_queued_node_messages(self.gateway_2, self.blockchain_fileno_2)

    def gateway_2_get_queued_messages_for_relay(self) -> List[AbstractMessage]:
        return helpers.get_queued_node_messages(self.gateway_2, self.relay_fileno_2)

    def gateway_2_get_queued_messages_for_gateway(self) -> List[AbstractMessage]:
        return helpers.get_queued_node_messages(self.gateway_2, self.gateway_fileno_2)



import datetime
import time

from bxcommon import constants
from bxcommon.connections.abstract_connection import AbstractConnection
from bxcommon.connections.connection_type import ConnectionType
from bxcommon.messages.bloxroute.broadcast_message import BroadcastMessage
from bxcommon.messages.bloxroute.key_message import KeyMessage
from bxcommon.utils import convert, crypto
from bxcommon.utils.object_hash import Sha256Hash
from bxcommon.utils.stats import stats_format
from bxcommon.utils.stats.block_stat_event_type import BlockStatEventType
from bxcommon.utils.stats.block_statistics_service import block_stats
from bxcommon.utils.stats.stat_block_type import StatBlockType
from bxgateway import gateway_constants
from bxgateway.gateway_constants import NeutralityPolicy
from bxgateway.messages.gateway.block_propagation_request import BlockPropagationRequestMessage
from bxutils import logging

logger = logging.get_logger(__name__)


class NeutralityService(object):
    """
    Service to manage block encryption and ensure network is neutral to Gateway Node's requests.
    """

    def __init__(self, node):
        self._node = node
        self._receipt_tracker = {}
        self._alarms = {}

    def register_for_block_receipts(self, cipher_hash, bx_block):
        """
        Register a block hash for receipts before broadcasting out key.
        :param cipher_hash: encrypted block ObjectHash
        :param bx_block compressed block
        """
        if cipher_hash in self._receipt_tracker:
            logger.debug("Ignoring duplicate bx_block hash for tracking receiving: {0}", cipher_hash)
            return

        self._receipt_tracker[cipher_hash] = 0
        if gateway_constants.NEUTRALITY_POLICY == NeutralityPolicy.RELEASE_IMMEDIATELY:
            logger.trace("Neutrality policy: releasing key immediately.")
            self._send_key(cipher_hash)
        else:
            logger.trace("Neutrality policy: waiting for receipts before releasing key.")
            alarm_id = self._node.alarm_queue.register_alarm(gateway_constants.NEUTRALITY_BROADCAST_BLOCK_TIMEOUT_S,
                                                             lambda: self._propagate_block_to_gateway_peers(cipher_hash,
                                                                                                            bx_block))
            self._alarms[cipher_hash] = alarm_id

    def record_block_receipt(self, cipher_hash, connection):
        """
        Records a receipt of a block hash. Releases key if threshold reached.
        :param cipher_hash encrypted block ObjectHash
        :param connection posting block received receipt
        """
        if cipher_hash in self._receipt_tracker:
            self._receipt_tracker[cipher_hash] += 1
            block_stats.add_block_event_by_block_hash(cipher_hash,
                                                      BlockStatEventType.ENC_BLOCK_RECEIVED_BLOCK_RECEIPT,
                                                      network_num=self._node.network_num,
                                                      peers=[connection],
                                                      more_info="{}, {} receipts".format(
                                                          stats_format.connection(connection),
                                                          self._receipt_tracker[cipher_hash]))

            if self._are_enough_receipts_received(cipher_hash):
                logger.debug("Received enough block receipt messages. Releasing key for block with hash: {}",
                             convert.bytes_to_hex(cipher_hash.binary))
                self._send_key(cipher_hash)
                self._node.alarm_queue.unregister_alarm(self._alarms[cipher_hash])
                del self._receipt_tracker[cipher_hash]
                del self._alarms[cipher_hash]

    def propagate_block_to_network(self, bx_block, connection, block_info=None, from_peer=False):
        """
        Propagates encrypted block to bloXroute network and starts listening for block receipts.
        :param bx_block: compressed block
        :param connection: connection initiating propagation
        :param block_info: original block hash, only provided if this is the original block
        :param from_peer: true if this message comes from another gateway. That means it is supposed to be encrypted
        :return: broadcast message
        """
        if self._node.opts.encrypt_blocks or from_peer:
            broadcast_msg = self._propagate_encrypted_block_to_network(bx_block, connection, block_info)
        else:
            broadcast_msg = self._propagate_unencrypted_block_to_network(bx_block, connection, block_info)
        return broadcast_msg

    def _propagate_encrypted_block_to_network(self, bx_block, connection, block_info):

        if block_info is None or block_info.block_hash is None:
            block_hash = b"Unknown"
            requested_by_peer = True
        else:
            block_hash = block_info.block_hash
            requested_by_peer = False

        encrypt_start_datetime = datetime.datetime.utcnow()
        encrypt_start_timestamp = time.time()
        encrypted_block, raw_cipher_hash = self._node.in_progress_blocks.encrypt_and_add_payload(bx_block)

        compressed_size = len(bx_block)
        encrypted_size = len(encrypted_block)

        encryption_details = "Encryption: {}; Size change: {}->{}bytes, {}".format(
            stats_format.timespan(encrypt_start_timestamp, time.time()),
            compressed_size, encrypted_size,
            stats_format.ratio(encrypted_size, compressed_size))

        block_stats.add_block_event_by_block_hash(block_hash,
                                                  BlockStatEventType.BLOCK_ENCRYPTED,
                                                  start_date_time=encrypt_start_datetime,
                                                  end_date_time=datetime.datetime.utcnow(),
                                                  network_num=self._node.network_num,
                                                  matching_block_hash=convert.bytes_to_hex(raw_cipher_hash),
                                                  matching_block_type=StatBlockType.ENCRYPTED.value,
                                                  more_info=encryption_details)

        cipher_hash = Sha256Hash(raw_cipher_hash)
        broadcast_message = BroadcastMessage(cipher_hash, self._node.network_num, is_encrypted=True,
                                             blob=encrypted_block)

        conns = self._node.broadcast(
            broadcast_message,
            connection,
            connection_types=(ConnectionType.RELAY_BLOCK,)
        )

        handling_duration = self._node.track_block_from_node_handling_ended(block_hash)
        block_stats.add_block_event_by_block_hash(cipher_hash,
                                                  BlockStatEventType.ENC_BLOCK_SENT_FROM_GATEWAY_TO_NETWORK,
                                                  network_num=self._node.network_num,
                                                  requested_by_peer=requested_by_peer,
                                                  peers=conns,
                                                  more_info="Peers: {}; {}; {}; Requested by peer: {}; Handled in {}"
                                                  .format(
                                                      stats_format.connections(conns), encryption_details,
                                                      self._format_block_info_stats(block_info), requested_by_peer,
                                                      stats_format.duration(handling_duration)))
        self.register_for_block_receipts(cipher_hash, bx_block)
        return broadcast_message

    def _propagate_unencrypted_block_to_network(self, bx_block, connection, block_info):
        if block_info is None:
            raise ValueError("Block info is required to propagate unencrypted block")

        broadcast_message = BroadcastMessage(block_info.block_hash, self._node.network_num, is_encrypted=False,
                                             blob=bx_block)
        conns = self._node.broadcast(
            broadcast_message,
            connection,
            connection_types=(ConnectionType.RELAY_BLOCK,)
        )
        handling_duration = self._node.track_block_from_node_handling_ended(block_info.block_hash)
        block_stats.add_block_event_by_block_hash(block_info.block_hash,
                                                  BlockStatEventType.ENC_BLOCK_SENT_FROM_GATEWAY_TO_NETWORK,
                                                  network_num=self._node.network_num,
                                                  requested_by_peer=False,
                                                  peers=conns,
                                                  more_info="Peers: {}; Unencrypted; {}; Handled in {}".format(
                                                      stats_format.connections(conns),
                                                      self._format_block_info_stats(block_info),
                                                      stats_format.duration(handling_duration)))
        logger.info("Propagating block {} to the BDN.", block_info.block_hash)
        return broadcast_message

    def _format_block_info_stats(self, block_info):
        if block_info is None:
            return ""

        return "Compression: {}, {}" \
            .format(stats_format.duration(block_info.duration_ms),
                    stats_format.percentage(block_info.compression_rate))

    def _are_enough_receipts_received(self, cipher_hash):
        neutrality_policy = gateway_constants.NEUTRALITY_POLICY
        receipt_count = self._receipt_tracker[cipher_hash]

        enough_by_count = receipt_count >= gateway_constants.NEUTRALITY_EXPECTED_RECEIPT_COUNT

        active_gateway_peer_count = len(
            list(
                filter(
                    lambda conn: conn.is_active(),
                    self._node.connection_pool.get_by_connection_types(
                        (ConnectionType.GATEWAY,)
                    )
                )
            )
        )
        if active_gateway_peer_count == 0:
            logger.debug("No active gateway peers to get block receipts from.")
            enough_by_percent = False
        else:
            enough_by_percent = (receipt_count / active_gateway_peer_count * 100 >=
                                 gateway_constants.NEUTRALITY_EXPECTED_RECEIPT_PERCENT)

        if neutrality_policy == NeutralityPolicy.RECEIPT_COUNT:
            return enough_by_count
        elif neutrality_policy == NeutralityPolicy.RECEIPT_PERCENT:
            return enough_by_percent
        elif neutrality_policy == NeutralityPolicy.RECEIPT_COUNT_AND_PERCENT:
            return enough_by_count and enough_by_percent

        raise ValueError("Unexpected neutrality policy: {}".format(neutrality_policy))

    def _propagate_block_to_gateway_peers(self, cipher_hash, bx_block):
        """
        Propagates unencrypted bx_block to all gateway peers for encryption and sending to bloXroute.
        Also sends keys to bloXroute in case this was user error (e.g. no gateway peers).
        Called after a timeout. This invalidates all future bx_block receipts.
        """
        bx_block_hash = crypto.double_sha256(bx_block)
        hex_bx_block_hash = convert.bytes_to_hex(bx_block_hash)

        logger.debug("Did not receive enough receipts for: {}. Propagating compressed block to other gateways: {}",
                     cipher_hash, hex_bx_block_hash)
        self._send_key(cipher_hash)

        request = BlockPropagationRequestMessage(bx_block)
        conns = self._node.broadcast(request, None, connection_types=(ConnectionType.GATEWAY,))
        block_stats.add_block_event_by_block_hash(cipher_hash,
                                                  BlockStatEventType.ENC_BLOCK_PROPAGATION_NEEDED,
                                                  network_num=self._node.network_num,
                                                  compressed_block_hash=hex_bx_block_hash,
                                                  peers=conns,
                                                  more_info="Peers: {}, {} receipts".format(
                                                      stats_format.connections(conns),
                                                      self._receipt_tracker[cipher_hash]))

        del self._receipt_tracker[cipher_hash]
        del self._alarms[cipher_hash]
        return constants.CANCEL_ALARMS

    def _send_key(self, cipher_hash):
        key = self._node.in_progress_blocks.get_encryption_key(bytes(cipher_hash.binary))
        key_message = KeyMessage(cipher_hash, self._node.network_num, key=key)
        conns = self._node.broadcast(
            key_message,
            None,
            connection_types=(ConnectionType.RELAY_BLOCK, ConnectionType.GATEWAY)
        )
        block_stats.add_block_event_by_block_hash(
            cipher_hash,
            BlockStatEventType.ENC_BLOCK_KEY_SENT_FROM_GATEWAY_TO_NETWORK,
            network_num=self._node.network_num,
            peers=conns,
        )



from dataclasses import dataclass, field
from typing import Dict
from typing import Optional

from bxcommon.models.config.abstract_config_model import AbstractConfigModel


@dataclass
class LogConfigModel(AbstractConfigModel):
    log_level: Optional[str] = None
    log_flush_immediately: Optional[bool] = None
    log_level_overrides: Dict[str, str] = field(default_factory=dict)



import blxr_rlp as rlp
from bxcommon.messages.eth.serializers.block_header import BlockHeader
from bxgateway.messages.eth.serializers.short_transaction import ShortTransaction


class CompactBlock(rlp.Serializable):
    fields = [
        ("header", BlockHeader),
        ("transactions", rlp.sedes.CountableList(ShortTransaction)),
        ("uncles", rlp.sedes.CountableList(BlockHeader)),
        ("chain_difficulty", rlp.sedes.big_endian_int),
        ("block_number", rlp.sedes.big_endian_int)
    ]



from typing import List

import blxr_rlp as rlp
from bxutils.logging.log_level import LogLevel

from bxgateway.messages.eth.protocol.eth_protocol_message import EthProtocolMessage
from bxgateway.messages.eth.protocol.eth_protocol_message_type import EthProtocolMessageType
from bxcommon.messages.eth.serializers.block_header import BlockHeader
from bxcommon.utils.blockchain_utils.eth import rlp_utils


class BlockHeadersEthProtocolMessage(EthProtocolMessage):
    msg_type = EthProtocolMessageType.BLOCK_HEADERS

    fields = [("block_headers", rlp.sedes.CountableList(BlockHeader))]

    def __repr__(self):
        headers = self.get_block_headers()
        headers_repr = list(headers[:1])
        if len(headers) > 1:
            headers_repr.append(headers[-1])
        return f"BlockHeadersEthProtocolMessage<headers_count: {len(headers)} " \
               f"headers: [{'...'.join([h.hash().hex() for h in headers_repr])}]>"

    def get_block_headers(self) -> List[BlockHeader]:
        return self.get_field_value("block_headers")

    def get_block_headers_bytes(self):
        if self._memory_view is None:
            self.serialize()

        return rlp_utils.get_first_list_field_items_bytes(self._memory_view)

    @classmethod
    def from_header_bytes(cls, header_bytes: memoryview) -> "BlockHeadersEthProtocolMessage":
        headers_list_prefix = rlp_utils.get_length_prefix_list(len(header_bytes))

        msg_bytes = bytearray(len(headers_list_prefix) + len(header_bytes))
        msg_bytes[:len(headers_list_prefix)] = headers_list_prefix
        msg_bytes[len(headers_list_prefix):] = header_bytes

        return cls(msg_bytes)

    def log_level(self):
        return LogLevel.DEBUG



class AbstractConfigModel:
    def merge(self, config: "AbstractConfigModel"):
        for key, value in self.__dict__.items():
            updated_attr = getattr(config, key, None)
            if updated_attr:
                if isinstance(updated_attr, dict) and hasattr(value, "__dict__"):
                    value.__dict__.update(updated_attr)
                else:
                    self.__dict__[key] = updated_attr
        return self



from abc import abstractmethod, ABC

from bxutils.logging.log_level import LogLevel


class AbstractMessage(ABC):
    HEADER_LENGTH = 0

    @classmethod
    @abstractmethod
    def unpack(cls, buf):
        """
        Unpack buffer into command, metadata (e.g. magic number, checksum), and payload.
        """

    @classmethod
    @abstractmethod
    def validate_payload(cls, buf, unpacked_args) -> None:
        """
        Validates unpacked content.
        """

    @classmethod
    @abstractmethod
    def initialize_class(cls, cls_type, buf, unpacked_args):
        """
        Initialize message class with arguments. Returns cls_type instance.
        """

    def log_level(self) -> LogLevel:
        return LogLevel.TRACE

    @abstractmethod
    def rawbytes(self) -> memoryview:
        """
        Returns raw message bytes of message.
        """



import struct
from typing import Optional

from bxgateway import ont_constants
from bxgateway.messages.ont.ont_message import OntMessage
from bxgateway.messages.ont.ont_message_type import OntMessageType
from bxutils.logging import LogLevel


class PingOntMessage(OntMessage):
    MESSAGE_TYPE = OntMessageType.PING

    def __init__(self, magic: Optional[int] = None, buf: Optional[bytearray] = None, height: Optional[int] = 0):
        if buf is None:
            buf = bytearray(ont_constants.ONT_HDR_COMMON_OFF + ont_constants.ONT_LONG_LONG_LEN)
            self.buf = buf

            off = ont_constants.ONT_HDR_COMMON_OFF
            struct.pack_into("<Q", buf, off, height)
            off += ont_constants.ONT_LONG_LONG_LEN

            super().__init__(magic, self.MESSAGE_TYPE, off - ont_constants.ONT_HDR_COMMON_OFF, buf)

        else:
            self.buf = buf
            self._memoryview = memoryview(buf)
            self._magic = self._command = self._payload_len = self._checksum = None
            self._payload = None
            self._height = height

    def height(self) -> int:
        if self._height == 0:
            if len(self.buf) == ont_constants.ONT_HDR_COMMON_OFF:
                self._height = -1
            else:
                self._height, = struct.unpack_from("<Q", self.buf, ont_constants.ONT_HDR_COMMON_OFF)

        return self._height



import struct
from typing import List, Tuple, Union

from bxcommon import constants


def get_serialized_length(count: int) -> int:
    return constants.UL_INT_SIZE_IN_BYTES + (count * constants.UL_INT_SIZE_IN_BYTES)


def serialize_short_ids_to_buffer(short_ids: List[int], buffer: bytearray, offset: int = 0) -> int:
    struct.pack_into("<L", buffer, offset, len(short_ids))
    offset += constants.UL_INT_SIZE_IN_BYTES

    for short_id in short_ids:
        struct.pack_into("<L", buffer, offset, short_id)
        offset += constants.UL_INT_SIZE_IN_BYTES

    return offset


def serialize_short_ids(short_ids: List[int]) -> bytearray:
    serialized_bytes = bytearray(get_serialized_length(len(short_ids)))
    serialize_short_ids_to_buffer(short_ids, serialized_bytes, 0)
    return serialized_bytes


def deserialize_short_ids_from_buffer(buffer: Union[bytearray, memoryview], offset: int = 0) -> Tuple[List[int], int]:
    short_ids = []

    short_ids_count, = struct.unpack_from("<L", buffer, offset)
    offset += constants.UL_INT_SIZE_IN_BYTES

    for _ in range(short_ids_count):
        short_id, = struct.unpack_from("<L", buffer, offset)
        short_ids.append(short_id)
        offset += constants.UL_INT_SIZE_IN_BYTES

    return short_ids, offset


def deserialize_short_ids(buffer: Union[bytearray, memoryview]) -> List[int]:
    short_ids, _ = deserialize_short_ids_from_buffer(buffer, 0)
    return short_ids



from dataclasses import dataclass

from bxcommon.models.serializeable_enum import SerializeableEnum


class NodeEventType(SerializeableEnum):
    PEER_CONN_ERR = "PEER_CONN_ERR"
    PEER_CONN_ESTABLISHED = "PEER_CONN_ESTABLISHED"
    PEER_CONN_CLOSED = "PEER_CONN_CLOSED"
    ONLINE = "ONLINE"
    OFFLINE = "OFFLINE"
    NOTIFY_OFFLINE = "NOTIFY_OFFLINE"
    SID_SPACE_EAGER_FETCH = "SID_SPACE_EAGER_FETCH"
    SID_SPACE_FULL = "SID_SPACE_FULL"
    SID_SPACE_SWITCH = "SID_SPACE_SWITCH"
    BLOCKCHAIN_NODE_CONN_ERR = "BLOCKCHAIN_NODE_CONN_ERR"
    BLOCKCHAIN_NODE_CONN_ESTABLISHED = "BLOCKCHAIN_NODE_CONN_ESTABLISHED"
    BLOCKCHAIN_NODE_CONN_ADDED = "BLOCKCHAIN_NODE_CONN_ADDED"
    BLOCKCHAIN_NODE_CONN_REMOVED = "BLOCKCHAIN_NODE_CONN_REMOVED"
    REMOTE_BLOCKCHAIN_CONN_ERR = "REMOTE_BLOCKCHAIN_CONN_ERR"
    REMOTE_BLOCKCHAIN_CONN_ESTABLISHED = "REMOTE_BLOCKCHAIN_CONN_ESTABLISHED"
    TX_SERVICE_FULLY_SYNCED = "TX_SERVICE_FULLY_SYNCED"
    TX_SERVICE_NOT_SYNCED = "TX_SERVICE_NOT_SYNCED"
    SWITCHING_RELAYS = "SWITCHING_RELAYS"

@dataclass
class NodeEventModel:
    # What event happened in a node.
    node_id: str
    # pyre-fixme[8]: Attribute has type `NodeEventType`; used as `None`.
    event_type: NodeEventType = None
    # pyre-fixme[8]: Attribute has type `str`; used as `None`.
    peer_ip: str = None
    # pyre-fixme[8]: Attribute has type `int`; used as `None`.
    peer_port: int = None
    # pyre-fixme[8]: Attribute has type `str`; used as `None`.
    timestamp: str = None
    # pyre-fixme[8]: Attribute has type `str`; used as `None`.
    event_id: str = None
    # pyre-fixme[8]: Attribute has type `str`; used as `None`.
    payload: str = None

    # 'type' has been deprecated but remains here for backwards compatibility with versions pre 1.54.0.
    # TODO: Remove "type" attribute and "__post_init__" once all gateway versions in the BDN are > v1.54.0. Also, remove
    #  the default value of "None" for "event_type"
    # pyre-fixme[8]: Attribute has type `NodeEventType`; used as `None`.
    type: NodeEventType = None

    def __post_init__(self):
        if self.event_type is None:
            self.event_type = self.type



from bxutils.log_message_categories import PROCESSING_FAILED_CATEGORY, GENERAL_CATEGORY, EXCEPTION_OR_CANCELLATION
from bxutils.logging_messages_utils import LogMessage

COULD_NOT_SERIALIZE_FEED_ENTRY = LogMessage(
    "C-000079",
    PROCESSING_FAILED_CATEGORY,
    "Could not serialize feed entry. Skipping."
)

BAD_FEED_SUBSCRIBER = LogMessage(
    "C-000068",
    GENERAL_CATEGORY,
    "Subscriber {} was not receiving messages and emptying its queue from "
    "{}. Disconnecting.",
)
COULD_NOT_DESERIALIZE_TRANSACTION = LogMessage(
    "C-000069",
    PROCESSING_FAILED_CATEGORY,
    "Could not deserialize transaction in transaction service to Ethereum payload: {}, body: {}. Error: {}",
)
BAD_FEED_SUBSCRIBER_SHOULD_EXIT = LogMessage(
    "C-000070",
    GENERAL_CATEGORY,
    "Subscriber {} was not filtering messages correctly from "
    "{}. Disconnecting.",
)
CONNECTION_DOES_NOT_EXIST = LogMessage(
    "C-000071",
    EXCEPTION_OR_CANCELLATION,
    "Connection doesn't exist, cancelling {} "
)



import struct
from typing import List, Optional

from bxcommon.constants import UL_INT_SIZE_IN_BYTES, CONTROL_FLAGS_LEN
from bxcommon.messages.bloxroute import blocks_short_ids_serializer
from bxcommon.messages.bloxroute.abstract_bloxroute_message import AbstractBloxrouteMessage
from bxcommon.messages.bloxroute.blocks_short_ids_serializer import BlockShortIds
from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType
from bxutils.logging.log_level import LogLevel


class TxServiceSyncBlocksShortIdsMessage(AbstractBloxrouteMessage):
    MESSAGE_TYPE = BloxrouteMessageType.TX_SERVICE_SYNC_BLOCKS_SHORT_IDS

    """
    Message used to send information about block short ids from requested relay
    """

    def __init__(
            self, network_num: Optional[int] = None, blocks_short_ids: Optional[List[BlockShortIds]] = None,
            buf: Optional[bytearray] = None
    ):
        # pyre-fixme[8]: Attribute has type `bytearray`; used as `Optional[bytearray]`.
        self.buf: bytearray = buf
        # pyre-fixme[8]: Attribute has type `int`; used as `Optional[int]`.
        self._network_num: int = network_num
        # pyre-fixme[8]: Attribute has type `int`; used as `Optional[int]`.
        self._block_count: int = len(blocks_short_ids) if blocks_short_ids is not None else None
        # pyre-fixme[8]: Attribute has type `List[BlockShortIds]`; used as
        #  `Optional[List[BlockShortIds]]`.
        self._blocks_short_ids: List[BlockShortIds] = blocks_short_ids

        if blocks_short_ids is not None and buf is None:
            self.buf = bytearray(
                self.HEADER_LENGTH +
                2 * UL_INT_SIZE_IN_BYTES
            )

            self._parse()

        self.buf.extend(bytearray(CONTROL_FLAGS_LEN))
        super(TxServiceSyncBlocksShortIdsMessage, self).__init__(
            self.MESSAGE_TYPE,
            len(self.buf) - self.HEADER_LENGTH,
            self.buf
        )

    def log_level(self) -> LogLevel:
        return LogLevel.DEBUG

    def network_num(self) -> int:
        if self._network_num is None:
            off = self.HEADER_LENGTH
            self._network_num, = struct.unpack_from("<L", self._memoryview, off)
        return self._network_num

    def block_count(self) -> int:
        if self._block_count is None:
            off = self.HEADER_LENGTH + UL_INT_SIZE_IN_BYTES
            self._block_count, = struct.unpack_from("<L", self._memoryview, off)
        return self._block_count

    def blocks_short_ids(self) -> List[BlockShortIds]:
        offset = self.HEADER_LENGTH + 2 * UL_INT_SIZE_IN_BYTES
        return blocks_short_ids_serializer.deserialize_blocks_short_ids_from_buffer(
            self._memoryview, offset, self.block_count()
        )

    def __repr__(self) -> str:
        return "{}<network_num: {}, block_count: {}".\
            format(self.__class__.__name__, self.network_num(), self.block_count())

    def _parse(self) -> None:
        off = self.HEADER_LENGTH
        struct.pack_into("<LL", self.buf, off, self._network_num, self._block_count)
        off += UL_INT_SIZE_IN_BYTES * 2

        self.buf.extend(blocks_short_ids_serializer.serialize_blocks_short_ids_into_bytes(self._blocks_short_ids))



from typing import List, Optional

from bxcommon.messages.bloxroute.abstract_cleanup_message import AbstractCleanupMessage
from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType
from bxcommon.utils import crypto
from bxcommon.utils.object_hash import Sha256Hash, NULL_SHA256_HASH


class TransactionCleanupMessage(AbstractCleanupMessage):
    MESSAGE_TYPE = BloxrouteMessageType.TRANSACTION_CLEANUP
    """
    Message with sids numbers for cleanup.
    """

    def __init__(self, network_num: Optional[int] = None, source_id: str = "", sids: Optional[List[int]] = None,
                 tx_hashes: Optional[List[Sha256Hash]] = None, buf: Optional[bytearray] = None):
        message_hash = NULL_SHA256_HASH
        build_buffer = buf is None

        super(TransactionCleanupMessage, self).__init__(message_hash, network_num, source_id, sids, tx_hashes, buf)
        if build_buffer:
            self.set_message_hash(
                Sha256Hash(crypto.double_sha256(self.buf[AbstractCleanupMessage.PAYLOAD_START_OFFSET:]))
            )

    def __repr__(self):
        return "TransactionCleanupMessage<message_hash: {}, network_num:{}, num_sids: {}, num_tx_hashes: {}>".format(
            self.message_hash(),
            self.network_num(),
            len(self.short_ids()),
            len(self.transaction_hashes())
        )



import struct
from typing import Optional

from bxcommon.utils.blockchain_utils.ont.ont_object_hash import OntObjectHash
from bxgateway import ont_constants
from bxgateway.messages.ont.ont_message import OntMessage


class DataOntMessage(OntMessage):
    def __init__(self, magic: Optional[int] = None, length: Optional[int] = None,
                 hash_start: Optional[OntObjectHash] = None, hash_stop: Optional[OntObjectHash] = None,
                 command: Optional[bytes] = None, buf: Optional[bytearray] = None):
        if buf is None:
            buf = bytearray(ont_constants.ONT_HDR_COMMON_OFF + ont_constants.ONT_DATA_MSG_LEN)
            self.buf = buf

            off = ont_constants.ONT_HDR_COMMON_OFF
            struct.pack_into("<B", buf, off, length)
            off += ont_constants.ONT_CHAR_LEN

            assert hash_start is not None
            assert hash_stop is not None
            buf[off:off + ont_constants.ONT_HASH_LEN] = hash_start.get_big_endian()
            off += ont_constants.ONT_HASH_LEN
            buf[off:off + ont_constants.ONT_HASH_LEN] = hash_stop.get_big_endian()
            off += ont_constants.ONT_HASH_LEN

            super().__init__(magic, command, off - ont_constants.ONT_HDR_COMMON_OFF, buf)

        else:
            self.buf = buf
            self._memoryview = memoryview(buf)
            self._magic = self._command = self._payload_len = self._checksum = None
            self._payload = None

        self._hash_start = self._hash_stop = None

    def hash_start(self) -> OntObjectHash:
        return OntObjectHash(
            buf=self.buf,
            offset=ont_constants.ONT_HDR_COMMON_OFF + self.payload_len() - 2 * ont_constants.ONT_HASH_LEN,
            length=ont_constants.ONT_HASH_LEN
        )

    def hash_stop(self) -> OntObjectHash:
        return OntObjectHash(
            buf=self.buf,
            offset=ont_constants.ONT_HDR_COMMON_OFF + self.payload_len() - ont_constants.ONT_HASH_LEN,
            length=ont_constants.ONT_HASH_LEN
        )



from typing import Dict, Optional

from bxutils.ssl.data.ssl_certificate_info import SSLCertificateInfo
from bxutils.ssl.ssl_certificate_type import SSLCertificateType


class SSLStorageInfo:
    ssl_folder_path: str
    certificates_info: Dict[SSLCertificateType, SSLCertificateInfo]

    def __init__(
            self,
            ssl_folder_path: str,
            ca_cert_info: SSLCertificateInfo,
            private_cert_info: SSLCertificateInfo,
            registration_cert_info: Optional[SSLCertificateInfo] = None,
     ):
        self.ssl_folder_path = ssl_folder_path
        self.certificates_info = {
            SSLCertificateType.CA: ca_cert_info,
            SSLCertificateType.PRIVATE: private_cert_info
        }
        if registration_cert_info is not None:
            self.certificates_info[SSLCertificateType.REGISTRATION_ONLY] = registration_cert_info

    def __repr__(self) -> str:
        return f"{self.__class__.__name__} <" \
            f"ssl_folder_path = {self.ssl_folder_path}, " \
            f"certificates_info = {self.certificates_info}" \
            f">"



from abc import ABCMeta, abstractmethod


class AbstractMessageConverterFactory:

    __metaclass__ = ABCMeta

    @abstractmethod
    def get_message_converter(self, msg_type):
        """
        Returns message converter for message type
        :param msg_type: message type
        :return: message converter
        """



import asyncio
from abc import abstractmethod, ABCMeta
from asyncio import Future, Task
from typing import Optional, Union, List, Any, Dict, Tuple, Callable, TypeVar, Coroutine

import websockets

from bxcommon.rpc.json_rpc_request import JsonRpcRequest
from bxcommon.rpc.json_rpc_response import JsonRpcResponse
from bxcommon.rpc.rpc_errors import RpcError, RpcTimedOut
from bxcommon import constants
from bxcommon.rpc.provider.abstract_provider import AbstractProvider, SubscriptionNotification
from bxcommon.rpc.provider.response_queue import ResponseQueue
from bxcommon.rpc.provider.subscription_manager import SubscriptionManager
from bxutils import log_messages
from bxutils import logging

T = TypeVar("T")
logger = logging.get_logger(__name__)


class WsException(Exception):
    # pylint: disable=super-init-not-called
    def __init__(self, message: str):
        self.message = message


class AbstractWsProvider(AbstractProvider, metaclass=ABCMeta):
    def __init__(
        self,
        uri: str,
        retry_connection: bool = False,
        queue_limit: int = constants.WS_PROVIDER_MAX_QUEUE_SIZE,
        headers: Optional[Dict] = None,
        retry_callback: Optional[Callable[["AbstractWsProvider"], Coroutine[None, None, None]]] = None
    ):
        self.uri = uri
        self.retry_connection = retry_connection
        self.retry_callback = retry_callback

        self.ws: Optional[websockets.WebSocketClientProtocol] = None
        self.listener_task: Optional[Future] = None
        self.ws_status_check: Optional[Future] = None
        self.connected_event = asyncio.Event()

        self.subscription_manager = SubscriptionManager(queue_limit)
        self.response_messages = ResponseQueue()
        self.tasks: List[Task] = []
        self.running = False
        self.current_request_id = 1
        if headers:
            self.headers = headers
        else:
            self.headers = {}

    async def initialize(self) -> None:
        await self.connect()

        # on initial connect, exception will be raised if self.ws is not set
        ws = self.ws
        assert ws is not None

    async def connect(self) -> None:
        logger.debug("Initiating websockets connection to: {}", self.uri)

        connection_attempts = 0
        while True:
            self.ws = None

            try:
                self.ws = await asyncio.wait_for(
                    self.connect_websocket(), constants.WS_MAX_CONNECTION_TIMEOUT_S
                )
            except (asyncio.TimeoutError, ConnectionRefusedError) as e:
                # immediately raise an error if this is the first attempt,
                # or not attempting to retry connections
                if not self.retry_connection or not self.running:
                    logger.warning(log_messages.WS_COULD_NOT_CONNECT, self.uri)
                    raise e
                if self.retry_connection:
                    logger.debug(
                        "Connection was rejected from websockets endpoint: {}. Attempts: {}. Retrying...",
                        self.uri,
                        connection_attempts
                    )
                    await asyncio.sleep(
                        constants.WS_RECONNECT_TIMEOUTS[connection_attempts]
                    )
                else:
                    logger.error(log_messages.WS_COULD_NOT_CONNECT_AFTER_RETRIES, self.uri, connection_attempts)
                    break
            # pylint: disable=broad-except
            except Exception as e:
                logger.error(log_messages.WS_UNEXPECTED_ERROR, e)
            else:
                break

            connection_attempts = min(connection_attempts, len(constants.WS_RECONNECT_TIMEOUTS) - 1)

        if self.ws is None:
            logger.debug("Could not reconnect websocket. Exiting.")
            self.running = False
        else:
            logger.debug("Connected to websockets endpoint: {}", self.uri)
            self.running = True
            ws_status_check = self.ws_status_check
            if ws_status_check is not None and not ws_status_check.done():
                ws_status_check.cancel()
            self.ws_status_check = asyncio.create_task(self._ensure_websocket_alive())

            self.connected_event.set()

            listener_task = self.listener_task
            if listener_task is not None and not listener_task.done():
                listener_task.cancel()
            self.listener_task = asyncio.create_task(self.receive())

    async def connect_websocket(self) -> websockets.WebSocketClientProtocol:
        return await websockets.connect(self.uri, extra_headers=self.headers, max_size=2**24)

    @abstractmethod
    async def subscribe(self, channel: str, options: Optional[Dict[str, Any]] = None) -> str:
        pass

    @abstractmethod
    async def unsubscribe(self, subscription_id: str) -> Tuple[bool, Optional[RpcError]]:
        pass

    async def call_rpc(
        self,
        method: str,
        params: Union[List[Any], Dict[Any, Any], None],
        request_id: Optional[str] = None
    ) -> JsonRpcResponse:
        if request_id is None:
            request_id = str(self.current_request_id)
            self.current_request_id += 1

        return await self.call(
            JsonRpcRequest(request_id, method, params)
        )

    async def call(
        self,
        request: JsonRpcRequest
    ) -> JsonRpcResponse:
        request_id = request.id
        assert request_id is not None

        await self.connected_event.wait()

        if not self.running:
            raise WsException(
                "Connection was broken when trying to call RPC method. "
                "Try reconnecting."
            )

        ws = self.ws
        assert ws is not None

        serialized_request = request.to_jsons()
        logger.trace("Sending message to websocket: {}", serialized_request)
        await ws.send(serialized_request)
        response = await self.get_rpc_response(request_id)
        if response is None:
            raise RpcTimedOut(None, "Please try again.")
        error = response.error
        if error:
            logger.error(
                log_messages.ETH_RPC_ERROR, error.message, error.data)
            raise error

        return response

    def subscribe_with_callback(
        self,
        callback: Callable[[SubscriptionNotification], None],
        channel: str,
        options: Optional[Dict[str, Any]] = None,
    ) -> Task:
        if options is None:
            options = {}
        subscribe_task = asyncio.create_task(
            self._handle_subscribe_callback(callback, channel, options)
        )
        self.tasks.append(subscribe_task)
        return subscribe_task

    async def _handle_subscribe_callback(
        self,
        callback: Callable[[SubscriptionNotification], None],
        channel: str,
        options: Dict[str, Any],
    ) -> None:
        subscription_id = await self.subscribe(channel, options)
        while self.running and subscription_id in self.subscription_manager.subscription_notifications_by_id:
            notification = await self.subscription_manager.get_next_subscription_notification_for_id(
                subscription_id
            )
            try:
                callback(notification)
            # pylint: disable=broad-except
            except Exception as e:
                logger.exception(log_messages.WS_COULD_NOT_PROCESS_NOTIFICATION, e)

    async def receive(self) -> None:
        # TODO: preferably this would be an infinite loop that waits on
        # self.connected_event.wait(), but it seems that `async for` never
        # throws an exception, even when the websocket gets disconnected.

        logger.trace("Started receiving on websocket.")
        ws = self.ws
        assert ws is not None

        async for next_message in ws:
            logger.trace("Received message on websocket: {}", next_message)

            # process response messages
            # noinspection PyBroadException
            try:
                response_message = JsonRpcResponse.from_jsons(next_message)
            # pylint: disable=broad-except
            except Exception:
                pass
            else:
                await self.response_messages.put(response_message)
                continue

            # process notification messages
            try:
                subscription_message = JsonRpcRequest.from_jsons(next_message)
                params = subscription_message.params
                assert isinstance(params, dict)
                await self.subscription_manager.receive_message(
                    SubscriptionNotification(
                        params["subscription"], params["result"],
                    )
                )
            # pylint: disable=broad-except
            except Exception as e:
                logger.warning(
                    log_messages.ETH_RPC_PROCESSING_ERROR,
                    next_message,
                    e,
                    exc_info=True
                )

        logger.trace("Temporarily stopped receiving message on websocket. Awaiting reconnection.")

    async def get_next_subscription_notification(self) -> SubscriptionNotification:
        task = asyncio.create_task(
            self.subscription_manager.get_next_subscription_notification()
        )
        return await self._wait_for_ws_message(task)

    async def get_next_subscription_notification_by_id_timeout(
        self, subscription_id: str
    ) -> SubscriptionNotification:
        task = asyncio.create_task(
            self.subscription_manager.get_next_subscription_notification_for_id(
                subscription_id
            )
        )
        return await self._wait_for_ws_message(task, timeout=constants.WAIT_FOR_SUBSCRIPTION_TIMEOUT)

    async def get_next_subscription_notification_by_id(
        self, subscription_id: str
    ) -> SubscriptionNotification:
        task = asyncio.create_task(
            self.subscription_manager.get_next_subscription_notification_for_id(
                subscription_id
            )
        )
        return await self._wait_for_ws_message(task)

    async def get_rpc_response(self, request_id: str) -> JsonRpcResponse:
        task = asyncio.create_task(self.response_messages.get_by_request_id(request_id))
        return await self._wait_for_ws_message(task)

    async def _wait_for_ws_message(
        self, ws_waiter_task: "Future[T]", timeout: Optional[int] = None
    ) -> T:
        if not self.connected_event.is_set():
            ws_waiter_task.cancel()
            raise WsException("Cannot wait for a message on a broken socket.")

        ws_status_check = self.ws_status_check
        assert ws_status_check is not None

        await asyncio.wait(
            [ws_waiter_task, ws_status_check], return_when=asyncio.FIRST_COMPLETED, timeout=timeout
        )

        if ws_waiter_task.done():
            return ws_waiter_task.result()
        else:
            ws_waiter_task.cancel()
            raise WsException(
                "Websocket connection disconnected while waiting for a response. "
                "Try reconnecting or checking SSL certificates."
            )

    async def _ensure_websocket_alive(self) -> None:
        ws = self.ws
        if ws is None:
            await asyncio.sleep(constants.WS_MIN_RECONNECT_TIMEOUT_S)
        else:
            await ws.wait_closed()

            if ws.transfer_data_exc:
                logger.error(
                    log_messages.RPC_TRANSPORT_EXCEPTION, ws.transfer_data_exc
                )

        self.connected_event.clear()

        if not self.retry_connection:
            listener = self.listener_task
            assert listener is not None
            listener.cancel()
            logger.debug("Websockets connection was broken. Closing...")
            self.running = False
            await self.close()
        elif self.running:
            logger.debug("Websockets connection was broken, reconnecting...")
            for task in self.tasks:
                task.cancel()
            self.tasks.clear()

            self.subscription_manager.subscription_notifications_by_id.clear()
            asyncio.create_task(self.reconnect())

    async def reconnect(self) -> None:
        await asyncio.sleep(constants.WS_MIN_RECONNECT_TIMEOUT_S)

        await self.connect()

        if self.ws is not None:
            logger.info("Websockets connection was re-established.")

            retry_callback = self.retry_callback
            if retry_callback:
                await retry_callback(self)

    async def close(self) -> None:
        logger.trace("Closing websockets provider")

        self.running = False

        ws = self.ws
        if ws is not None:
            await ws.close()
            await ws.wait_closed()

        listener = self.listener_task
        if listener is not None:
            listener.cancel()

        ws_status_check = self.ws_status_check
        if ws_status_check is not None:
            ws_status_check.cancel()

        self.connected_event.set()



class BtcMessageType(object):
    VERSION = b"version"
    VERACK = b"verack"
    PING = b"ping"
    PONG = b"pong"
    GET_ADDRESS = b"getaddr"
    ADDRESS = b"addr"
    INVENTORY = b"inv"
    GET_DATA = b"getdata"
    NOT_FOUND = b"notfound"
    GET_HEADERS = b"getheaders"
    GET_BLOCKS = b"getblocks"
    TRANSACTIONS = b"tx"
    BLOCK = b"block"
    HEADERS = b"headers"
    REJECT = b"reject"
    SEND_HEADERS = b"sendheaders"
    COMPACT_BLOCK = b"cmpctblock"
    SEND_COMPACT = b"sendcmpct"
    GET_BLOCK_TRANSACTIONS = b"getblocktxn"
    BLOCK_TRANSACTIONS = b"blocktxn"
    FEE_FILTER = b"feefilter"
    XVERSION = b"xversion"



import struct

from bxcommon import constants
from bxcommon.constants import DEFAULT_NETWORK_NUM
from bxcommon.messages.bloxroute import compact_block_short_ids_serializer
from bxcommon.messages.bloxroute.bloxroute_version_manager import bloxroute_version_manager
from bxcommon.messages.bloxroute.broadcast_message import BroadcastMessage
from bxcommon.messages.bloxroute.hello_message import HelloMessage
from bxcommon.messages.bloxroute.key_message import KeyMessage
from bxcommon.messages.bloxroute.tx_message import TxMessage
from bxcommon.models.broadcast_message_type import BroadcastMessageType
from bxcommon.test_utils import helpers
from bxcommon.utils import crypto
from bxcommon.utils.object_hash import Sha256Hash


def hello_message():
    return HelloMessage(protocol_version=bloxroute_version_manager.CURRENT_PROTOCOL_VERSION,
                        network_num=DEFAULT_NETWORK_NUM)


def broadcast_key_pair(short_ids=None, network_num=0):
    if short_ids is None:
        short_ids = [1, 10, 99, 187]
    broadcast_message_hash = Sha256Hash(helpers.generate_bytearray(crypto.SHA256_HASH_LEN))
    broadcast_message_bytes = bytearray(constants.UL_ULL_SIZE_IN_BYTES)
    broadcast_message_bytes.extend(helpers.generate_bytearray(500))
    struct.pack_into("<Q", broadcast_message_bytes, 0, len(broadcast_message_bytes))
    broadcast_message_bytes.extend(compact_block_short_ids_serializer.serialize_short_ids_into_bytes(short_ids))
    key_bytes, enc_broadcast_message_bytes = crypto.symmetric_encrypt(bytes(broadcast_message_bytes))

    broadcast_message = BroadcastMessage(broadcast_message_hash, network_num, "", BroadcastMessageType.BLOCK, True,
                                         enc_broadcast_message_bytes)
    key_message = KeyMessage(broadcast_message_hash, network_num, "", key_bytes)
    return broadcast_message, key_message


def tx_message(
    network_num: int = 1,
    source_id: str = "",
    short_id: int = constants.NULL_TX_SID
) -> TxMessage:
    return TxMessage(
        helpers.generate_object_hash(),
        network_num,
        source_id,
        short_id,
        helpers.generate_bytearray(250),
        account_id=helpers.generate_object_hash().to_string()
    )



from bxcommon import constants
from bxcommon.messages.abstract_internal_message import AbstractInternalMessage
from bxcommon.messages.bloxroute.abstract_bloxroute_message import AbstractBloxrouteMessage
from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType
from bxcommon.messages.bloxroute.v13.pong_message_v13 import PongMessageV13
from bxcommon.messages.bloxroute.pong_message import PongMessage
from bxcommon.messages.versioning.abstract_message_converter import AbstractMessageConverter


class _PongMessageConverterV13(AbstractMessageConverter):
    _MSG_TYPE_TO_OLD_MSG_CLASS_MAPPING = {
        BloxrouteMessageType.PONG: PongMessageV13
    }

    _MSG_TYPE_TO_NEW_MSG_CLASS_MAPPING = {
        BloxrouteMessageType.PONG: PongMessage
    }

    _BASE_LENGTH = (
        AbstractBloxrouteMessage.HEADER_LENGTH
    )

    _BREAKPOINT = (
        _BASE_LENGTH + constants.UL_ULL_SIZE_IN_BYTES
    )

    _OLD_MESSAGE_LEN = (
        _BASE_LENGTH + PongMessageV13.KEEP_ALIVE_MESSAGE_LENGTH
    )

    _NEW_MESSAGE_LEN = (
        _BASE_LENGTH + PongMessage.PONG_MESSAGE_LENGTH
    )

    _LENGTH_DIFFERENCE = (
        constants.UL_ULL_SIZE_IN_BYTES
    )

    def convert_to_older_version(
        self, msg: AbstractInternalMessage
    ) -> AbstractInternalMessage:
        msg_type = msg.MESSAGE_TYPE

        if msg_type not in self._MSG_TYPE_TO_OLD_MSG_CLASS_MAPPING:
            raise ValueError(
                f"Tried to convert unexpected new message type to v10: {msg_type}"
            )

        old_version_msg_class = self._MSG_TYPE_TO_OLD_MSG_CLASS_MAPPING[
            msg_type
        ]
        old_version_payload_len = msg.payload_len() - self._LENGTH_DIFFERENCE

        old_version_msg_bytes = bytearray(self._OLD_MESSAGE_LEN)
        old_version_msg_bytes[:self._BREAKPOINT] = msg.rawbytes()[:self._BREAKPOINT]
        old_version_msg_bytes[self._BREAKPOINT:] = msg.rawbytes()[self._BREAKPOINT + self._LENGTH_DIFFERENCE:]

        return AbstractBloxrouteMessage.initialize_class(
            old_version_msg_class,
            old_version_msg_bytes,
            (msg_type, old_version_payload_len),
        )

    def convert_from_older_version(
        self, msg: AbstractInternalMessage
    ) -> AbstractInternalMessage:
        msg_type = msg.MESSAGE_TYPE

        if msg_type not in self._MSG_TYPE_TO_NEW_MSG_CLASS_MAPPING:
            raise ValueError(
                f"Tried to convert unexpected old message type from v10: {msg_type}"
            )

        new_msg_class = self._MSG_TYPE_TO_NEW_MSG_CLASS_MAPPING[msg_type]
        new_payload_len = msg.payload_len() + self._LENGTH_DIFFERENCE

        new_msg_bytes = bytearray(self._NEW_MESSAGE_LEN)

        new_msg_bytes[:self._BREAKPOINT] = msg.rawbytes()[:self._BREAKPOINT]

        new_msg_bytes[self._BREAKPOINT + self._LENGTH_DIFFERENCE:] = msg.rawbytes()[self._BREAKPOINT:]

        return AbstractBloxrouteMessage.initialize_class(
            new_msg_class,
            new_msg_bytes,
            (msg_type, new_payload_len)
        )

    def convert_first_bytes_to_older_version(
        self, first_msg_bytes: memoryview
    ) -> memoryview:
        raise NotImplementedError

    def convert_first_bytes_from_older_version(
        self, first_msg_bytes: memoryview
    ) -> memoryview:
        raise NotImplementedError

    def convert_last_bytes_to_older_version(
        self, last_msg_bytes: memoryview
    ) -> memoryview:
        raise NotImplementedError

    def convert_last_bytes_from_older_version(
        self, last_msg_bytes: memoryview
    ) -> memoryview:
        raise NotImplementedError

    def get_message_size_change_to_older_version(self) -> int:
        return -self._LENGTH_DIFFERENCE

    def get_message_size_change_from_older_version(self) -> int:
        return self._LENGTH_DIFFERENCE


pong_message_converter_v13 = _PongMessageConverterV13()



from collections import deque
from typing import TypeVar, Generic, Dict, Deque, Optional

KT = TypeVar("KT")
VT = TypeVar("VT")


class LimitedSizeDict(Generic[KT, VT]):
    """
    Dictionary with expiration time.
    """

    contents: Dict[KT, VT]
    key_tracker: Deque[KT]
    _max_size: int

    def __init__(self, max_size: int) -> None:
        self.contents = {}
        self.key_tracker = deque()
        self._max_size = max_size

    def __contains__(self, item: KT) -> bool:
        return item in self.contents

    def __setitem__(self, key: KT, value: VT) -> None:
        if key in self.contents:
            self.contents[key] = value
        else:
            self.add(key, value)

    def __delitem__(self, key: KT) -> None:
        """
        Deletion is fairly expensive. Generally, there is no need to delete
        from this data structure.
        """
        del self.contents[key]
        self.key_tracker.remove(key)

    def __getitem__(self, item: KT) -> VT:
        return self.contents[item]

    def __len__(self) -> int:
        return len(self.contents)

    def add(self, key: KT, value: VT) -> None:
        self._make_room_for_new_item()

        self.contents[key] = value
        self.key_tracker.append(key)

    def get(self, key: KT, default: Optional[VT] = None) -> Optional[VT]:
        return self.contents.get(key, default)

    def _make_room_for_new_item(self) -> None:
        while len(self.key_tracker) >= self._max_size:
            key = self.key_tracker.popleft()
            del self.contents[key]



import time
from ssl import SSLContext
from typing import List, Optional, Dict, Any, cast

from bxcommon.constants import SdnRoutes
from bxcommon.models.blockchain_network_model import BlockchainNetworkModel
from bxcommon.models.gateway_settings_model import GatewaySettingsModel
from bxcommon.models.node_event_model import NodeEventModel, NodeEventType
from bxcommon.models.node_model import NodeModel
from bxcommon.models.outbound_peer_model import OutboundPeerModel
from bxcommon.models.bdn_account_model_base import BdnAccountModelBase
from bxcommon.services import http_service
from bxcommon.utils import model_loader, ip_resolver
from bxutils import log_messages
from bxutils import logging
from bxutils.encoding import json_encoder

logger = logging.get_logger(__name__)


def fetch_node_attributes(node_id: str) -> Optional[NodeModel]:
    # Should only be used for test networks.
    node_url = SdnRoutes.node.format(node_id)
    opts = cast(Dict[str, Any], http_service.get_json(node_url))
    logger.trace("Retrieved config for id {0} : {1}", node_id, opts)

    if opts:
        return model_loader.load_model(NodeModel, opts)
    else:
        return None


def _fetch_peers(node_url: str, node_id: Optional[str] = None) -> List[OutboundPeerModel]:
    outbound_peers_response = cast(List[Dict[str, Any]], http_service.get_json(node_url))
    logger.debug(
        "Retrieved outbound peers for node {0} from endpoint {1}: {2}",
        node_id,
        node_url,
        outbound_peers_response
    )

    if not outbound_peers_response:
        logger.warning(log_messages.BDN_RETURNED_NO_PEERS, node_url)
        return []

    outbound_peers = [
        model_loader.load_model(OutboundPeerModel, o) for o in outbound_peers_response
    ]
    ip_resolver.blocking_resolve_peers(outbound_peers)
    return outbound_peers


def fetch_potential_relay_peers_by_network(node_id: str, network_num: int) -> Optional[List[OutboundPeerModel]]:
    node_url = SdnRoutes.node_potential_relays_by_network.format(node_id, network_num)
    return _fetch_peers(node_url, node_id)


def fetch_gateway_peers(node_id: str, request_streaming: bool) -> Optional[List[OutboundPeerModel]]:
    node_url = SdnRoutes.node_gateways.format(node_id, request_streaming)
    return _fetch_peers(node_url, node_id)


def fetch_remote_blockchain_peer(node_id: str) -> Optional[OutboundPeerModel]:
    node_url = SdnRoutes.node_remote_blockchain.format(node_id)
    peers = _fetch_peers(node_url, node_id)
    if len(peers) < 1:
        logger.warning(log_messages.BDN_RETURNED_UNEXPECTED_NUMBER_OF_PEERS)
        return None
    else:
        logger.debug("Ordered potential remote blockchain peers: {}", peers)
        return peers[0]


def fetch_blockchain_network(protocol_name: str, network_name: str) -> Optional[BlockchainNetworkModel]:
    node_url = SdnRoutes.blockchain_network.format(protocol_name, network_name)
    blockchain_network = cast(Dict[str, Any], http_service.get_json(node_url))

    if blockchain_network is None:
        return None

    blockchain_network = model_loader.load_model(BlockchainNetworkModel, blockchain_network)

    return blockchain_network


def fetch_gateway_settings(node_id: str) -> GatewaySettingsModel:
    node_url = SdnRoutes.gateway_settings.format(node_id)
    gateway_settings = cast(Dict[str, Any], http_service.get_json(node_url))

    if not gateway_settings:
        gateway_settings = GatewaySettingsModel()
    else:
        gateway_settings = model_loader.load_model(GatewaySettingsModel, gateway_settings)

    return gateway_settings


def fetch_blockchain_networks() -> Dict[int, BlockchainNetworkModel]:
    node_url = SdnRoutes.blockchain_networks
    sdn_blockchain_networks = http_service.get_json(node_url)

    if not sdn_blockchain_networks:
        logger.warning(log_messages.BDN_CONTAINS_NO_CONFIGURED_NETWORKS)
        return {}

    assert isinstance(sdn_blockchain_networks, List)
    blockchain_network_models = [model_loader.load_model(BlockchainNetworkModel, b) for b in sdn_blockchain_networks]
    blockchain_network_dict = {network_model.network_num: network_model for network_model in blockchain_network_models}
    return blockchain_network_dict


def fetch_quota_status(account_id: str) -> Optional[Dict[str, Any]]:
    result = http_service.get_json_with_payload(SdnRoutes.quota_status, {"account_id": account_id})
    if result:
        result = cast(Dict[str, Any], result)
        return result
    else:
        return None


def submit_sid_space_switch(node_id: str) -> None:
    submit_node_event(NodeEventModel(node_id=node_id, event_type=NodeEventType.SID_SPACE_SWITCH))


def submit_sid_space_full_event(node_id: str) -> None:
    submit_node_event(NodeEventModel(node_id=node_id, event_type=NodeEventType.SID_SPACE_FULL))


def submit_peer_connection_error_event(node_id: str, peer_ip: str, peer_port: int) -> None:
    submit_peer_connection_event(NodeEventType.PEER_CONN_ERR, node_id, peer_ip, peer_port)


def submit_peer_connection_event(
    event_type: NodeEventType, node_id: str, peer_ip: str, peer_port: int, payload: Optional[str] = None
) -> None:
    submit_node_event(
        # pyre-fixme[6]: Expected `str` for 5th param but got `Optional[str]`.
        NodeEventModel(node_id=node_id, event_type=event_type, peer_ip=peer_ip, peer_port=peer_port, payload=payload)
    )


def submit_gateway_switching_relays_event(node_id: str) -> None:
    submit_node_event(NodeEventModel(node_id=node_id, event_type=NodeEventType.SWITCHING_RELAYS))


def submit_gateway_inbound_connection(node_id: str, peer_id: str) -> None:
    http_service.post_json(SdnRoutes.gateway_inbound_connection.format(node_id), peer_id)


def submit_tx_synced_event(node_id: str) -> None:
    submit_node_event(NodeEventModel(node_id=node_id, event_type=NodeEventType.TX_SERVICE_FULLY_SYNCED))


def submit_tx_not_synced_event(node_id: str) -> None:
    submit_node_event(NodeEventModel(node_id=node_id, event_type=NodeEventType.TX_SERVICE_NOT_SYNCED))


def submit_notify_offline_event(node_id: str) -> None:
    submit_node_event(NodeEventModel(node_id=node_id, event_type=NodeEventType.NOTIFY_OFFLINE))


def delete_gateway_inbound_connection(node_id: str, peer_id: str) -> None:
    http_service.delete_json(SdnRoutes.gateway_inbound_connection.format(node_id), peer_id)


def submit_node_event(node_event_model: NodeEventModel) -> None:
    node_event_model.timestamp = str(time.time())
    logger.trace("Submitting event for node {0} {1}", node_event_model.node_id, json_encoder.to_json(node_event_model))
    url = SdnRoutes.node_event.format(node_event_model.node_id)
    http_service.post_json(url, node_event_model)


def register_node(node_model: NodeModel) -> NodeModel:
    if not node_model:
        raise ValueError("Missing node model.")

    # Let the SDN know who this is.
    # SDN determines peers and returns full config.
    node_config = cast(Dict[str, Any], http_service.post_json(SdnRoutes.nodes, node_model))

    logger.trace("Registered node. Response: {}", json_encoder.to_json(node_config))

    if not node_config:
        raise EnvironmentError("Unable to reach SDN and register this node. Please check connection.")

    registered_node_model = model_loader.load_model(NodeModel, node_config)

    if not registered_node_model.source_version:
        raise ValueError(f"Source version {node_model.source_version} is no longer supported. Please upgrade to the "
                         f"latest version")

    if registered_node_model.blockchain_network_num == -1:
        raise ValueError(f"The blockchain network number {node_model.blockchain_network_num} does not exists. Please "
                         f"check the blockchain network startup parameters")

    return registered_node_model


def reset_pool(ssl_context: SSLContext) -> None:
    http_service.update_http_ssl_context(ssl_context)


def fetch_account_model(account_id: str) -> Optional[BdnAccountModelBase]:
    account_url = SdnRoutes.account.format(account_id)
    response = cast(Dict[str, Any], http_service.get_json(account_url))

    if response:
        return model_loader.load_model(BdnAccountModelBase, response)
    else:
        return None



from typing import Optional

from bxcommon import constants
from bxcommon.messages.bloxroute.bloxroute_message_control_flags import BloxrouteMessageControlFlags
from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType
from bxcommon.messages.validation.abstract_message_validator import AbstractMessageValidator
from bxcommon.messages.validation.control_flag_validation_error import ControlFlagValidationError
from bxcommon.messages.validation.message_size_validation_settings import \
    MessageSizeValidationSettings
from bxcommon.messages.validation.message_validation_error import MessageValidationError
from bxcommon.utils import convert
from bxcommon.utils.buffers.input_buffer import InputBuffer


class BloxrouteMessageValidator(AbstractMessageValidator):
    FIRST_VALIDATING_VERSION = 4

    def __init__(
        self,
        size_validation_settings: Optional[MessageSizeValidationSettings],
        connection_protocol_version: int
    ):
        self._size_validation_settings: Optional[MessageSizeValidationSettings] = size_validation_settings
        self._connection_protocol_version: int = connection_protocol_version

    def validate(
        self,
        is_full_msg: bool,
        msg_type: Optional[bytes],
        header_len: int,
        payload_len: Optional[int],
        input_buffer: InputBuffer
    ) -> None:
        """
        Validates message payload length.
        Throws MessageValidationError is message is not valid

        :param is_full_msg: indicates if the full message is available on input buffer
        :param msg_type: message type
        :param header_len: message header length
        :param payload_len: message payload length
        :param input_buffer: input buffer
        """

        if self._connection_protocol_version >= self.FIRST_VALIDATING_VERSION:
            self._validate_starting_sequence(input_buffer)

        if self._size_validation_settings is not None:
            self._validate_payload_length(msg_type, payload_len)

        if self._connection_protocol_version >= self.FIRST_VALIDATING_VERSION:
            self._validate_control_flags(is_full_msg, header_len, payload_len, input_buffer)

    def _validate_starting_sequence(self, input_buffer: InputBuffer) -> None:

        if input_buffer.length < constants.STARTING_SEQUENCE_BYTES_LEN:
            return

        starting_sequence_bytes = input_buffer[:constants.STARTING_SEQUENCE_BYTES_LEN]
        if starting_sequence_bytes != constants.STARTING_SEQUENCE_BYTES:
            raise MessageValidationError(
                f"Expected message to begin with starting sequence "
                f"but received first bytes "
                f"'{convert.bytes_to_hex(starting_sequence_bytes)}'"
            )

    def _validate_payload_length(self, msg_type: Optional[bytes], payload_len: Optional[int]) -> None:
        if msg_type is None or payload_len is None:
            return

        if msg_type == BloxrouteMessageType.TRANSACTION:
            size_validation_settings = self._size_validation_settings
            assert size_validation_settings is not None
            if payload_len > size_validation_settings.max_tx_size_bytes:
                raise MessageValidationError(
                    f"Transaction message size exceeds expected max size. "
                    f"Expected: {size_validation_settings.max_tx_size_bytes}. "
                    f"Actual: {payload_len}."
                )

        elif msg_type in {
            BloxrouteMessageType.BROADCAST,
            BloxrouteMessageType.TRANSACTIONS,
            BloxrouteMessageType.TX_SERVICE_SYNC_BLOCKS_SHORT_IDS,
            BloxrouteMessageType.TX_SERVICE_SYNC_TXS,
            BloxrouteMessageType.TRANSACTION_CLEANUP,
            BloxrouteMessageType.BLOCK_CONFIRMATION,
        }:
            size_validation_settings = self._size_validation_settings
            assert size_validation_settings is not None
            if payload_len > size_validation_settings.max_block_size_bytes:
                raise MessageValidationError(
                    f"{msg_type} message size exceeds expected max size. "
                    f"Expected: {size_validation_settings.max_block_size_bytes}. "
                    f"Actual: {payload_len}."
                )

        elif payload_len > constants.DEFAULT_MAX_PAYLOAD_LEN_BYTES:
            raise MessageValidationError(
                f"Message by type '{msg_type}' exceeds expected payload len. "
                f"Expected: {constants.DEFAULT_MAX_PAYLOAD_LEN_BYTES}. "
                f"Actual: {payload_len}."
            )

    def _validate_control_flags(
        self,
        is_full: bool,
        header_len: int,
        payload_len: Optional[int],
        input_buffer: InputBuffer
    ) -> None:
        if not is_full:
            return

        assert payload_len is not None
        if input_buffer.length < header_len + payload_len:
            raise MessageValidationError(
                f"Not enough bytes in the input buffer to get control flags. "
                f"Header length: {header_len}. "
                f"Payload length: {payload_len}. "
                f"Input buffer length: {input_buffer.length}"
            )

        control_flag_byte = input_buffer[header_len + payload_len - 1:header_len + payload_len]
        if BloxrouteMessageControlFlags.VALID not in BloxrouteMessageControlFlags(control_flag_byte[0]):
            raise ControlFlagValidationError(
                f"Control flags byte does not have VALID flag set. Value: {control_flag_byte}.",
                control_flag_byte[0]
            )



import asyncio
import functools
import signal
import socket
from asyncio import CancelledError, Future
from asyncio.events import AbstractServer
from ssl import SSLContext
from typing import Iterator, List, Coroutine, Generator, Callable, Awaitable, Optional

from bxcommon import constants
from bxcommon.connections.abstract_node import AbstractNode
from bxcommon.connections.connection_state import ConnectionState
from bxcommon.connections.connection_type import ConnectionType
from bxcommon.exceptions import HighMemoryError, FeedSubscriptionTimeoutError
from bxcommon.network.base_socket_connection_protocol import BaseSocketConnectionProtocol
from bxcommon.network.ip_endpoint import IpEndpoint
from bxcommon.network.peer_info import ConnectionPeerInfo
from bxcommon.network.transport_layer_protocol import TransportLayerProtocol
from bxcommon.utils.expiring_dict import ExpiringDict
from bxutils import logging, constants as utils_constants

logger = logging.get_logger(__name__)


async def run_until_other_completed(
    callback: Callable[..., Awaitable], tsk_future: Future, wait_future: Future
) -> None:
    done = False
    future_to_set: Optional[Future] = tsk_future
    while not done:
        await callback()
        done = wait_future.done()
        if future_to_set is not None:
            future_to_set.set_result(True)
            future_to_set = None


class NodeEventLoop:
    _node: AbstractNode
    _stop_requested: bool

    def __init__(self, node: AbstractNode) -> None:
        self._node = node
        self._stop_requested = False
        loop = asyncio.get_event_loop()
        self._started = loop.create_future()
        loop.add_signal_handler(signal.SIGTERM, self.stop)
        loop.add_signal_handler(signal.SIGINT, self.stop)
        loop.add_signal_handler(signal.SIGSEGV, self.stop)

        self.connected_peers = ExpiringDict(
            self._node.alarm_queue,
            constants.THROTTLE_RECONNECT_TIME_S,
            name="throttle_reconnect"
        )

    async def run(self) -> None:
        try:
            await self._run()
        except HighMemoryError as e:
            raise e
        except FeedSubscriptionTimeoutError as e:
            raise e
        except Exception as e: # pylint: disable=broad-except
            logger.exception("Unhandled error raised: {}.", e)

    async def close(self) -> None:
        await self._node.close()

    def stop(self) -> None:
        logger.info("Stopping node event loop due to a termination request.")
        self._node.should_force_exit = True
        self._stop_requested = True

    async def wait_started(self) -> None:
        if self._started is not None:
            await self._started
            self._started = None

    async def _run(self) -> None:
        node_servers = await self._create_servers()
        try:
            await self._node.init()
            loop = asyncio.get_event_loop()
            await self._connect_to_peers()
            self._node.fire_alarms()
            self._started.set_result(True)
            while not self._stop_requested:
                node_future = loop.create_future()
                connection_future = loop.create_future()
                await asyncio.gather(
                    run_until_other_completed(self._perform_node_tasks, node_future, connection_future),
                    run_until_other_completed(self._process_new_connections_requests, connection_future, node_future)
                )
                if self._node.force_exit():
                    logger.info("Ending event loop. Shutdown has been requested.")
                    break
        finally:
            await self.close()
            for server in node_servers:
                server.close()
        if self._node.should_restart_on_high_memory:
            raise HighMemoryError()

    async def _process_new_connections_requests(self) -> None:
        peers_info = self._node.dequeue_connection_requests()
        if peers_info is not None:
            await asyncio.gather(*self._gather_connections(iter(peers_info)))
        else:
            await asyncio.sleep(constants.MAX_EVENT_LOOP_TIMEOUT)

    async def _connect_to_peers(self) -> None:
        connection_futures = self._gather_connections(self._iter_outbound_peers())
        if connection_futures:
            await asyncio.gather(*connection_futures)

    async def _create_servers(self) -> List[AbstractServer]:
        loop = asyncio.get_event_loop()
        endpoints = self._node.server_endpoints
        server_futures = []
        for endpoint in endpoints:
            if not endpoint.port:
                logger.debug("Endpoint: {} ,port is required, skipping", endpoint)
                continue
            ssl_ctx = None
            if endpoint.port in utils_constants.SSL_PORT_RANGE:
                ssl_ctx = self._node.get_server_ssl_ctx()
            server_future = loop.create_server(
                functools.partial(self._protocol_factory, endpoint, True),
                endpoint.ip_address,
                endpoint.port,
                family=socket.AF_INET,
                backlog=constants.DEFAULT_NODE_BACKLOG,
                reuse_address=True,
                ssl=ssl_ctx
            )
            server_futures.append(server_future)
        logger.debug("Starting listening on: {}.", endpoints)
        # pyre-fixme[7]: Expected `List[AbstractServer]` but got `Tuple[typing.Any]`.
        return await asyncio.gather(*server_futures)

    async def _connect_to_target(
        self,
        peer_info: ConnectionPeerInfo
    ) -> None:
        logger.debug("Connecting to {}.", peer_info)
        loop = asyncio.get_event_loop()
        target_endpoint = peer_info.endpoint
        try:
            if peer_info.transport_protocol == TransportLayerProtocol.TCP:
                ssl_ctx = self._get_target_ssl_context(target_endpoint, peer_info.connection_type)
                conn_task = asyncio.ensure_future(loop.create_connection(
                    functools.partial(self._protocol_factory, target_endpoint),
                    target_endpoint.ip_address,
                    target_endpoint.port,
                    family=socket.AF_INET,
                    ssl=ssl_ctx
                ))
            else:
                conn_task = asyncio.ensure_future(loop.create_datagram_endpoint(
                    functools.partial(self._protocol_factory, target_endpoint, is_ssl=False),
                    remote_addr=(target_endpoint.ip_address, target_endpoint.port),
                    family=socket.AF_INET
                ))
            await asyncio.wait_for(conn_task, constants.CONNECTION_TIMEOUT)
        except (
            TimeoutError,
            asyncio.TimeoutError,
            CancelledError,
            ConnectionRefusedError,
            ConnectionResetError,
            OSError,
        ) as e:
            err = str(e)
            if not err:
                err = repr(e)

            self._node.log_refused_connection(peer_info, err)
            self._node.handle_connection_closed(True, peer_info, ConnectionState.ESTABLISHED)

    def _iter_outbound_peers(self) -> Generator[ConnectionPeerInfo, None, None]:
        sdn_address = self._node.get_sdn_address()
        if sdn_address:
            yield ConnectionPeerInfo(IpEndpoint(*sdn_address), ConnectionType.SDN)
        else:
            logger.debug("SDN address not provided, skipping connection. This is expected for gateways.")
        for peer_info in self._node.get_outbound_peer_info():
            yield peer_info

    def _gather_connections(
        self, connections_info: Iterator[ConnectionPeerInfo]
    ) -> List[Coroutine]:
        return [
            self._connect_to_target(peer_info) for peer_info in connections_info
        ]

    def _get_target_ssl_context(self, endpoint: IpEndpoint, connection_type: ConnectionType) -> Optional[SSLContext]:
        if endpoint.port in utils_constants.SSL_PORT_RANGE:
            return self._node.get_target_ssl_ctx(endpoint, connection_type)
        else:
            return None

    async def _perform_node_tasks(self) -> None:
        timeout = self._node.fire_alarms()
        self._node.flush_all_send_buffers()
        if timeout is None or timeout < 0:
            timeout = constants.MAX_EVENT_LOOP_TIMEOUT
        else:
            timeout = min(timeout, constants.MAX_EVENT_LOOP_TIMEOUT)
        await asyncio.sleep(timeout)

    def _protocol_factory(
        self,
        endpoint: IpEndpoint,
        is_server: bool = False,
        is_ssl: Optional[bool] = None
    ) -> BaseSocketConnectionProtocol:
        if is_server:
            target_endpoint = None
        else:
            target_endpoint = endpoint

        if is_ssl is None:
            is_ssl = endpoint.port in utils_constants.SSL_PORT_RANGE

        return BaseSocketConnectionProtocol(self._node, target_endpoint, is_ssl=is_ssl)



from typing import Optional

from bxgateway import ont_constants
from bxgateway.messages.ont.ont_message import OntMessage
from bxgateway.messages.ont.ont_message_type import OntMessageType


class GetAddrOntMessage(OntMessage):
    MESSAGE_TYPE = OntMessageType.GET_ADDRESS

    def __init__(self, magic: Optional[int] = None, buf: Optional[bytearray] = None):
        if buf is None:
            buf = bytearray(ont_constants.ONT_HDR_COMMON_OFF)
            self.buf = buf

            super().__init__(magic, self.MESSAGE_TYPE, 0, buf)
        else:
            self.buf = buf
            self._memoryview = memoryview(buf)
            self._magic = self._command = self._payload_len = self._checksum = None
            self._payload = None



from bxgateway.messages.eth.abstract_eth_message import AbstractEthMessage


class EthProtocolMessage(AbstractEthMessage):
    def __init__(self, msg_bytes, *args, **kwargs):
        super(EthProtocolMessage, self).__init__(msg_bytes, *args, **kwargs)

    def __repr__(self):
        return "EthProtocolMessage<type: {}>".format(self.__class__.__name__)

    @classmethod
    def unpack(cls, buf):
        pass

    @classmethod
    def validate_payload(cls, buf, unpacked_args):
        pass




import asyncio
from asyncio import Future
from typing import Optional, cast, List, Dict, Any, TYPE_CHECKING

from bxcommon.connections.connection_type import ConnectionType
from bxcommon.exceptions import FeedSubscriptionTimeoutError
from bxcommon.feed.feed import FeedKey
from bxcommon.messages.bloxroute.abstract_bloxroute_message import AbstractBloxrouteMessage
from bxcommon.models.transaction_key import TransactionKey
from bxcommon.rpc.external.eth_ws_subscriber import EthWsSubscriber
from bxcommon.rpc.provider.abstract_ws_provider import WsException
from bxcommon.services.transaction_service import TransactionService
from bxcommon.utils import convert
from bxcommon.utils.object_hash import Sha256Hash
from bxcommon.feed.feed_manager import FeedManager
from bxcommon.feed.feed_source import FeedSource
from bxgateway import log_messages
from bxcommon.feed.eth.eth_raw_transaction import EthRawTransaction
from bxcommon.feed.eth.eth_pending_transaction_feed import EthPendingTransactionFeed
from bxgateway.connections.gateway_connection import GatewayConnection
from bxgateway.messages.gateway.confirmed_tx_message import ConfirmedTxMessage
from bxgateway.utils.stats.transaction_feed_stats_service import transaction_feed_stats_service
from bxutils import logging

if TYPE_CHECKING:
    from bxgateway.connections.eth.eth_gateway_node import EthGatewayNode

logger = logging.get_logger(__name__)

SUBSCRIBE_REQUEST_ID = "1"


class EthWsProxyPublisher(EthWsSubscriber):
    """
    Publishes transactions accepted to Ethereum mempool to a `pendingTxs` feed.
    """

    def __init__(
        self,
        ws_uri: str,
        feed_manager: FeedManager,
        transaction_service: TransactionService,
        node: "EthGatewayNode"
    ) -> None:
        self.feed_manager = feed_manager
        self.transaction_service = transaction_service
        self.node = node

        # ok, lifecycle patterns are a bit different
        super().__init__(ws_uri)
        self.receiving_tasks: List[Future] = []
        self.stream_confirmation_messages = node.opts.stream_to_peer_gateway is not None

    async def revive(self) -> None:
        """
        Revives subscriber; presumably, subscriber got disconnected earlier
        and stopped retrying.
        """
        if self.ws is None and not self.running:
            logger.info("Attempting to revive Ethereum websockets feed...")
            await self.reconnect()

    async def reconnect(self) -> None:
        logger.warning(log_messages.ETH_WS_SUBSCRIBER_CONNECTION_BROKEN)

        for receiving_task in self.receiving_tasks:
            receiving_task.cancel()
        self.receiving_tasks = []

        try:
            await super().reconnect()
        except ConnectionRefusedError:
            self.running = False

        if self.running:
            await self.subscribe_to_feeds()
            logger.info("Reconnected to Ethereum websocket feed")
        else:
            logger.warning(log_messages.ETH_RPC_COULD_NOT_RECONNECT)

    async def start(self) -> None:
        ws_uri = self.uri
        if ws_uri is not None:

            await self.initialize()

            logger.info("Subscribed to Ethereum websocket feed.")
            await self.subscribe_to_feeds()

    async def subscribe_to_feeds(self):
        subscription_id = await self.subscribe("newPendingTransactions")
        self.receiving_tasks.append(asyncio.create_task(self.handle_tx_notifications(subscription_id)))

        subscription_id = await self.subscribe("newHeads")
        self.receiving_tasks.append(asyncio.create_task(self.handle_block_notifications(subscription_id)))

    async def handle_tx_notifications(self, subscription_id: str) -> None:
        while self.running:
            next_notification = await self.get_next_subscription_notification_by_id(
                subscription_id
            )
            transaction_hash = next_notification.notification
            assert isinstance(transaction_hash, str)
            self.process_received_transaction(
                self.transaction_service.get_transaction_key(
                    Sha256Hash(
                        convert.hex_to_bytes(
                            transaction_hash[2:]
                        )
                    )
                )
            )

    async def handle_block_notifications(self, subscription_id: str) -> None:
        while self.running:
            next_notification = await self.get_next_subscription_notification_by_id_timeout(
                subscription_id
            )
            if not next_notification:
                raise FeedSubscriptionTimeoutError()
            logger.debug(
                "NewBlockHeader Notification {} from node", next_notification
            )
            block_header = next_notification.notification
            block_hash = Sha256Hash(convert.hex_to_bytes(block_header["hash"][2:]))
            block_number = int(block_header["number"], 16)
            block_difficulty = int(block_header["difficulty"], 16)

            self.node.publish_block(
                block_number, block_hash, None, FeedSource.BLOCKCHAIN_RPC
            )
            self.node.block_processing_service.set_last_confirmed_block_parameters(
                block_number, block_difficulty
            )

    def process_received_transaction(self, transaction_key: TransactionKey) -> None:
        tx_contents = cast(
            Optional[memoryview],
            self.transaction_service.get_transaction_by_key(transaction_key)
        )
        if tx_contents is None:
            asyncio.create_task(self.fetch_missing_transaction(transaction_key))
        else:
            self.process_transaction_with_contents(transaction_key, tx_contents)

    def process_transaction_with_contents(
        self, transaction_key: TransactionKey, tx_contents: memoryview
    ) -> None:
        transaction_feed_stats_service.log_pending_transaction_from_local(transaction_key.transaction_hash)

        self.feed_manager.publish_to_feed(
            FeedKey(EthPendingTransactionFeed.NAME, self.node.network_num),
            EthRawTransaction(
                transaction_key.transaction_hash,
                tx_contents,
                FeedSource.BLOCKCHAIN_RPC,
                local_region=True
            )
        )
        if self.stream_confirmation_messages:
            self.broadcast_confirmation_message(ConfirmedTxMessage(transaction_key.transaction_hash))

    async def fetch_missing_transaction(self, transaction_key: TransactionKey) -> None:
        try:
            response = await self.call_rpc(
                "eth_getTransactionByHash",
                [f"0x{str(transaction_key.transaction_hash)}"]
            )
            self.process_transaction_with_parsed_contents(transaction_key, response.result)
        except WsException:
            # ok, don't continue processing
            logger.debug(
                "Attempt to fetch transaction {} was interrupted by a broken connection. "
                "Abandoning.",
                transaction_key.transaction_hash
            )
            pass

    def process_transaction_with_parsed_contents(
        self, transaction_key: TransactionKey, parsed_tx: Optional[Dict[str, Any]]
    ) -> None:
        transaction_feed_stats_service.log_pending_transaction_from_local(transaction_key.transaction_hash)

        if parsed_tx is None:
            logger.debug(log_messages.TRANSACTION_NOT_FOUND_IN_MEMPOOL, transaction_key.transaction_hash)
            transaction_feed_stats_service.log_pending_transaction_missing_contents()
        else:
            gas_price = int(parsed_tx["gasPrice"], 16)
            if gas_price >= self.node.get_network_min_transaction_fee():
                self.feed_manager.publish_to_feed(
                    FeedKey(EthPendingTransactionFeed.NAME, network_num=self.node.network_num),
                    EthRawTransaction(
                        transaction_key.transaction_hash,
                        parsed_tx,
                        FeedSource.BLOCKCHAIN_RPC,
                        local_region=True,
                    )
                )

            if self.stream_confirmation_messages:
                self.broadcast_confirmation_message(ConfirmedTxMessage(transaction_key.transaction_hash))

    def broadcast_confirmation_message(self, message: AbstractBloxrouteMessage) -> None:
        gateway_connections = cast(
            List[GatewayConnection],
            self.node.connection_pool.get_by_connection_types((ConnectionType.EXTERNAL_GATEWAY,))
        )
        for connection in gateway_connections:
            if connection.stream_confirmation_messages:
                connection.enqueue_msg(message)

    async def stop(self) -> None:
        await self.close()
        for receiving_task in self.receiving_tasks:
            receiving_task.cancel()



import random
import struct
import sys

from bxutils.logging.log_level import LogLevel

from bxgateway.btc_constants import BTC_HDR_COMMON_OFF
from bxgateway.messages.btc.btc_message import BtcMessage
# FIXME dedup this against pongbtcmessage
from bxgateway.messages.btc.btc_message_type import BtcMessageType


class PingBtcMessage(BtcMessage):
    MESSAGE_TYPE = BtcMessageType.PING

    def __init__(self, magic=None, buf=None):
        if buf is None:
            buf = bytearray(BTC_HDR_COMMON_OFF + 8)
            self.buf = buf

            off = BTC_HDR_COMMON_OFF
            struct.pack_into("<Q", buf, off, random.randint(0, sys.maxsize))
            off += 8

            BtcMessage.__init__(self, magic, self.MESSAGE_TYPE, off - BTC_HDR_COMMON_OFF, buf)
        else:
            self.buf = buf
            self._memoryview = memoryview(buf)
            self._magic = self._command = self._payload_len = self._checksum = None
            self._payload = None

        self._nonce = None

    def nonce(self):
        if self._nonce is None:
            if len(self.buf) == BTC_HDR_COMMON_OFF:
                self._nonce = -1
            elif len(self.buf) == BTC_HDR_COMMON_OFF + 4:
                self._nonce = struct.unpack_from("<L", self.buf, BTC_HDR_COMMON_OFF)[0]
            else:
                self._nonce = struct.unpack_from("<Q", self.buf, BTC_HDR_COMMON_OFF)[0]

        return self._nonce

    def log_level(self) -> LogLevel:
        return LogLevel.DEBUG



from typing import NamedTuple, Dict, Any

from bxcommon import constants
from bxcommon.rpc.rpc_errors import RpcInvalidParams
from bxcommon.utils import convert
from bxcommon.utils.object_hash import Sha256Hash
from bxcommon.rpc import rpc_constants
from bxcommon.feed.feed import Feed
from bxcommon.feed.subscriber import Subscriber
from bxcommon.feed.feed_source import FeedSource
from bxutils import logging

logger = logging.get_logger(__name__)


class RawTransactionFeedEntry:
    tx_hash: str
    tx_contents: str
    local_region: bool

    def __init__(
        self,
        tx_hash: Sha256Hash,
        tx_contents: memoryview,
        local_region: bool
    ) -> None:
        self.tx_hash = str(tx_hash)
        self.tx_contents = convert.bytes_to_hex(tx_contents)
        self.local_region = local_region

    def __eq__(self, other) -> bool:
        return (
            isinstance(other, RawTransactionFeedEntry)
            and other.tx_hash == self.tx_hash
            and other.tx_contents == self.tx_contents
            and other.local_region == self.local_region
        )


class RawTransaction(NamedTuple):
    tx_hash: Sha256Hash
    tx_contents: memoryview
    source: FeedSource
    local_region: bool


class NewTransactionFeed(Feed[RawTransactionFeedEntry, RawTransaction]):
    NAME = rpc_constants.NEW_TRANSACTION_FEED_NAME
    FIELDS = ["tx_hash", "tx_contents", "local_region"]
    ALL_FIELDS = FIELDS

    def __init__(self, network_num: int = constants.ALL_NETWORK_NUM,) -> None:
        super().__init__(self.NAME, network_num)

    def subscribe(
        self, options: Dict[str, Any]
    ) -> Subscriber[RawTransactionFeedEntry]:
        include_from_blockchain = options.get("include_from_blockchain", None)
        if include_from_blockchain is not None:
            if not isinstance(include_from_blockchain, bool):
                raise RpcInvalidParams(
                    "\"include_from_blockchain\" must be a boolean"
                )
        return super().subscribe(options)

    def serialize(self, raw_message: RawTransaction) -> RawTransactionFeedEntry:
        return RawTransactionFeedEntry(
            raw_message.tx_hash, raw_message.tx_contents, raw_message.local_region
        )

    # pylint: disable=unused-argument
    def should_publish_message_to_subscriber(
        self,
        subscriber: Subscriber[RawTransactionFeedEntry],
        raw_message: RawTransaction,
        serialized_message: RawTransactionFeedEntry
    ) -> bool:
        if (
            raw_message.source == FeedSource.BLOCKCHAIN_SOCKET
            and not subscriber.options.get("include_from_blockchain", False)
        ):
            return False
        else:
            return True



import hashlib
import os
import struct

import bitcoin
from Crypto.Hash import keccak
from coincurve import PrivateKey, PublicKey

from bxcommon.utils import convert
from bxcommon.utils.blockchain_utils.eth import eth_common_utils, eth_common_constants
from bxcommon.utils.blockchain_utils.eth import rlp_utils


def sign(msg, private_key):
    """
    Signs message using private key
    """

    if not msg:
        raise ValueError("Message is required")

    if len(private_key) != eth_common_constants.PRIVATE_KEY_LEN:
        raise ValueError("Private key is expected of len {0} but was {1}"
                         .format(eth_common_constants.PRIVATE_KEY_LEN, len(private_key)))

    new_private_key = PrivateKey(private_key)
    return new_private_key.sign_recoverable(msg, hasher=None)


def get_sha3_calculator(input_bytes):
    """
    Returns object that can be used to calculate sha3 hash
    :param input_bytes: input bytes
    :return: object that can calculate sha3 256 hash
    """

    if input_bytes is None:
        raise ValueError("Input is required")

    return keccak.new(digest_bits=eth_common_constants.SHA3_LEN_BITS, update_after_digest=True, data=input_bytes)


def recover_public_key(message, signature, hasher=None):
    """
    Recovers public key from signed message
    :param message: message
    :param signature: signature
    :param hasher: hash function to use on message (usually sha256 or keccak_hash)
    :return: public key
    """

    if len(signature) != eth_common_constants.SIGNATURE_LEN:
        raise ValueError("Expected signature len of {0} but was {1}"
                         .format(eth_common_constants.SIGNATURE_LEN, len(signature)))

    public_key = PublicKey.from_signature_and_message(signature, message, hasher=hasher)
    return public_key.format(compressed=False)[1:]


def verify_signature(pubkey, signature, message):
    """
    Verifies signature
    :param pubkey: signing public key
    :param signature: message signature
    :param message: signature
    :return: returns True if signature is valid, False otherwise
    """

    if len(pubkey) != eth_common_constants.PUBLIC_KEY_LEN:
        raise ValueError("Pubkey is expected of len {0} but was {1}"
                         .format(eth_common_constants.PUBLIC_KEY_LEN, len(pubkey)))

    if len(signature) != eth_common_constants.SIGNATURE_LEN:
        raise ValueError("Signature is expected of len {0} but was {1}"
                         .format(eth_common_constants.PUBLIC_KEY_LEN, len(signature)))

    if not message:
        raise ValueError("Message is required")

    public_key = PublicKey.from_signature_and_message(signature, message, hasher=None)
    return public_key.format(compressed=False) == b"\04" + pubkey


def encode_signature(tx_v: int, tx_r: int, tx_s: int) -> bytes:
    """
    Calculates byte representation of ECC signature from parameters
    :param tx_v:
    :param tx_r:
    :param tx_s:
    :return: bytes of ECC signature
    """
    if tx_v > eth_common_constants.EIP155_CHAIN_ID_OFFSET:
        if tx_v % 2 == 0:
            tx_v = 28
        else:
            tx_v = 27

    if tx_v not in (27, 28):
        raise ValueError("v is expected to be int or long in range (27, 28)")

    return encode_signature_y_parity(tx_v - 27, tx_r, tx_s)


def encode_signature_y_parity(y_parity: int, tx_r: int, tx_s: int) -> bytes:
    v_bytes = rlp_utils.ascii_chr(y_parity)
    # pyre-fixme[16]: Module `bitcoin` has no attribute `encode`.
    # pyre-fixme[16]: Module `bitcoin` has no attribute `encode`.
    r_bytes, s_bytes = bitcoin.encode(tx_r, 256), bitcoin.encode(tx_s, 256)
    return _left_0_pad_32(r_bytes) + _left_0_pad_32(s_bytes) + v_bytes


def decode_signature(sig):
    """
    Decodes coordinates from ECC signature bytes
    :param sig: signature bytes
    :return: ECC signature parameters
    """

    if not sig:
        raise ValueError("Signature is required")

    return rlp_utils.safe_ord(sig[64]) + 27, bitcoin.decode(sig[0:32], 256), bitcoin.decode(sig[32:64], 256)


def make_private_key(seed):
    """
    Generates ECC private using provided seed value
    :param seed: seed used to generate ECC private key
    :return: ECC private key
    """

    if not seed:
        raise ValueError("Seed is required")

    return eth_common_utils.keccak_hash(seed)


def generate_random_private_key_hex_str():
    """
    Generate hex string of random ECC private key
    :return: ECC private key hex string
    """

    # seed can be any random bytes of any length
    random_seed = os.urandom(100)
    private_key_bytes = make_private_key(random_seed)
    return convert.bytes_to_hex(private_key_bytes)


def private_to_public_key(raw_private_key):
    """
    Calculates public key for private key
    :param raw_private_key: ECC private key
    :return: public key
    """

    if len(raw_private_key) != eth_common_constants.PRIVATE_KEY_LEN:
        raise ValueError("Private key is expected of len {0} but was {1}"
                         .format(eth_common_constants.PRIVATE_KEY_LEN, len(raw_private_key)))

    raw_pubkey = bitcoin.encode_pubkey(bitcoin.privtopub(raw_private_key), "bin_electrum")
    assert len(raw_pubkey) == eth_common_constants.PUBLIC_KEY_LEN
    return raw_pubkey


def ecies_kdf(key_material, key_len):
    """
    interop w/go ecies implementation

    for sha3, blocksize is 136 bytes
    for sha256, blocksize is 64 bytes

    NIST SP 800-56a Concatenation Key Derivation Function (see section 5.8.1).

    :param key_material: key material
    :param key_len: key length
    :return: key
    """

    if len(key_material) != eth_common_constants.KEY_MATERIAL_LEN:
        raise ValueError("Key material is expected of len {0} but was {1}"
                         .format(eth_common_constants.KEY_MATERIAL_LEN, len(key_material)))

    if key_len <= 0:
        raise ValueError("Key len is expected to be positive but was {0}".format(key_len))

    empty_bytes = b""
    key = b""
    hash_blocksize = eth_common_constants.BLOCK_HASH_LEN
    reps = ((key_len + 7) * 8) / (hash_blocksize * 8)
    counter = 0
    while counter <= reps:
        counter += 1
        ctx = hashlib.sha256()
        ctx.update(struct.pack(">I", counter))
        ctx.update(key_material)
        ctx.update(empty_bytes)
        key += ctx.digest()
    return key[:key_len]


def string_xor(string_1, string_2):
    """
    Calculates xor of two strings
    :param string_1: string 1
    :param string_2: string 2
    :return: xor of two strings
    """

    if len(string_1) != len(string_2):
        raise ValueError("String must have the same length")

    return b"".join(
        rlp_utils.ascii_chr(rlp_utils.safe_ord(a) ^ rlp_utils.safe_ord(b)) for a, b in zip(string_1, string_2)
    )


def get_padded_len_16(input_bytes):
    """
    Length of bytes if padded to 16
    :param input_bytes: bytes
    :return: padded length
    """

    return input_bytes if input_bytes % eth_common_constants.MSG_PADDING == 0 \
        else input_bytes + eth_common_constants.MSG_PADDING - (input_bytes % eth_common_constants.MSG_PADDING)


def right_0_pad_16(data):
    """
    Pads bytes with 0 on the right to length of 16
    :param data: bytes
    :return: padded bytes
    """

    if len(data) % eth_common_constants.MSG_PADDING:
        data += b"\x00" * (eth_common_constants.MSG_PADDING - len(data) % eth_common_constants.MSG_PADDING)
    return data


def _left_0_pad_32(input_bytes):
    """
    Pads bytes with 0 on the left to length of 32
    :param input_bytes: bytes
    :return: padded bytes
    """

    return b"\x00" * (32 - len(input_bytes)) + input_bytes


def public_key_to_address(public_key_bytes: bytes) -> bytes:
    return eth_common_utils.keccak_hash(memoryview(public_key_bytes))[-20:]



from enum import Enum
from typing import Union

from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType
from bxcommon.utils.object_hash import Sha256Hash

from bxgateway.messages.btc.btc_message_type import BtcMessageType
from bxgateway.messages.eth.protocol.eth_protocol_message_type import EthProtocolMessageType


class MessageConversionType(Enum):
    BLOCK_COMPRESSION = "BlockCompression"
    BLOCK_DECOMPRESSION = "BlockDecompression"
    COMPACT_BLOCK_COMPRESSION = "CompactBlockCompression"


class MessageConversionError(Exception):

    def __init__(
            self,
            msg_hash: Sha256Hash,
            src_msg_type: bytes,
            target_msg_type: bytes,
            error: str,
            conversion_type: "MessageConversionType"
    ):
        self.error_msg = f"failed to convert {src_msg_type.decode()} to {target_msg_type.decode()} - {error}"


        self.msg_hash = msg_hash
        self.conversion_type = conversion_type
        super(MessageConversionError, self).__init__(self.error_msg)


def btc_block_decompression_error(msg_hash: Sha256Hash, error: Union[Exception, str]) -> "MessageConversionError":
    return MessageConversionError(
        msg_hash, BloxrouteMessageType.BROADCAST, BtcMessageType.BLOCK, str(error), MessageConversionType.BLOCK_DECOMPRESSION
    )


def btc_block_compression_error(msg_hash: Sha256Hash, error: Union[Exception, str]) -> "MessageConversionError":
    return MessageConversionError(
        msg_hash, BtcMessageType.BLOCK, BloxrouteMessageType.BROADCAST, str(error), MessageConversionType.BLOCK_COMPRESSION
    )


def eth_block_compression_error(msg_hash: Sha256Hash, error: Union[Exception, str]) -> "MessageConversionError":
    return MessageConversionError(
        msg_hash, EthProtocolMessageType.NEW_BLOCK_BYTES, BloxrouteMessageType.BROADCAST, str(error), MessageConversionType.BLOCK_COMPRESSION
    )


def eth_block_decompression_error(msg_hash: Sha256Hash, error: Union[Exception, str]) -> "MessageConversionError":
    return MessageConversionError(
        msg_hash, BloxrouteMessageType.BROADCAST, EthProtocolMessageType.NEW_BLOCK_BYTES, str(error), MessageConversionType.BLOCK_DECOMPRESSION
    )


def btc_compact_block_compression_error(msg_hash: Sha256Hash, error: Union[Exception, str]) -> "MessageConversionError":
    return MessageConversionError(
        msg_hash,
        BtcMessageType.COMPACT_BLOCK,
        BloxrouteMessageType.BROADCAST,
        str(error),
        MessageConversionType.COMPACT_BLOCK_COMPRESSION
    )



import time

from bxcommon.messages.bloxroute.block_confirmation_message import BlockConfirmationMessage
from bxcommon.services.transaction_service import TransactionService
from bxcommon.utils.blockchain_utils.ont.ont_object_hash import OntObjectHash
from bxgateway.messages.ont.block_ont_message import BlockOntMessage
from bxcommon.utils import crypto
from bxgateway.services.ont.abstract_ont_block_cleanup_service import AbstractOntBlockCleanupService
from bxutils import logging
from bxutils.logging.log_record_type import LogRecordType
from bxgateway import ont_constants

from bxcommon.services import normal_cleanup_service_helpers

logger = logging.get_logger(LogRecordType.BlockCleanup)


class OntNormalBlockCleanupService(AbstractOntBlockCleanupService):
    """
    Service for managing block cleanup.
    """

    # TODO: Implement block cleanup
    def clean_block_transactions(self, block_msg: BlockOntMessage, transaction_service: TransactionService) -> None:
        block_short_ids = []
        block_unknown_tx_hashes = []
        start_time = time.time()

        short_ids_count = 0
        unknown_tx_hashes_count = 0
        transactions_processed = 0

        tx_hash_to_contents_len_before_cleanup = transaction_service.get_tx_hash_to_contents_len()
        short_id_count_before_cleanup = transaction_service.get_short_id_count()

        for tx in block_msg.txns():
            tx_hash = OntObjectHash(
                buf=crypto.double_sha256(tx),
                length=ont_constants.ONT_HASH_LEN
            )
            short_ids = transaction_service.remove_transaction_by_tx_hash(tx_hash, force=True)
            if short_ids is None:
                unknown_tx_hashes_count += 1
                block_unknown_tx_hashes.append(tx_hash)
            else:
                short_ids_count += len(short_ids)
                block_short_ids.extend(short_ids)
            transactions_processed += 1
        block_hash = block_msg.block_hash()
        transaction_service.on_block_cleaned_up(block_hash)
        end_time = time.time()
        duration = end_time - start_time
        tx_hash_to_contents_len_after_cleanup = transaction_service.get_tx_hash_to_contents_len()
        short_id_count_after_cleanup = transaction_service.get_short_id_count()

        logger.debug(
            "Finished cleaning up block {}. Processed {} hashes, {} of which were unknown, and cleaned up {} "
            "short ids. Took {:.3f}s.",
            block_hash, transactions_processed, unknown_tx_hashes_count, short_ids_count, duration
        )

        transaction_service.log_block_transaction_cleanup_stats(block_hash, block_msg.txn_count(),
                                                                tx_hash_to_contents_len_before_cleanup,
                                                                tx_hash_to_contents_len_after_cleanup,
                                                                short_id_count_before_cleanup,
                                                                short_id_count_after_cleanup)

        self._block_hash_marked_for_cleanup.discard(block_hash)
        self.node.post_block_cleanup_tasks(
            block_hash=block_hash,
            short_ids=block_short_ids,
            unknown_tx_hashes=block_unknown_tx_hashes
        )

    # pyre-fixme[14]: `contents_cleanup` overrides method defined in
    #  `AbstractBlockCleanupService` inconsistently.
    def contents_cleanup(self, transaction_service: TransactionService,
                         block_confirmation_message: BlockConfirmationMessage):
        normal_cleanup_service_helpers.contents_cleanup(transaction_service, block_confirmation_message)
